"""
maclandrol: 22/07/19
This is an attempt to reverse engineer the BRICS (Breaking of Retrosynthetically Interesting Chemical Substructures)
approach for molecule fragmentation and use it as an heuristic for assembling molecules. The original paper on BRICS can be found here:
http://dacemirror.sci-hub.tw/journal-article/93060992e8d889318b77b562c0e5b75f/degen2008.pdf.
This makes senses from a methodological point of view, but I can't either guarantee that its is working as expected of if it's the best
way to tackle this problem. The goal here is to reconstruct a set of original molecules, which if they were to be fragmented using BRICS, should yield
the same fragment set in input. Thus, in theory fragments obtained using BRICS CAN be assembled into the original molecules with this method.
This differs from rdkit BRICSBuild implementation that requires the presence of dummy indicator atoms added by a prior BRICS fragmentation.
"""

from typing import Optional

import copy
import json
import itertools
import random
import re
import pkg_resources

from functools import lru_cache

import numpy as np

from rdkit import Chem
from rdkit.Chem import rdChemReactions

import datamol as dm
from ..types import Mol

CCQ = "[$([#6][!#6;!#1]):1]!@[#6;!a:2]>>[*:1].[*:2]"
CCQ_RETRO = "[$([#6;!H0][!#6;!#1]):1].[#6;!a;!H0:2]>>[*:1][*:2]"

BRICS_ENVIRONS = {
    "L1": "[C;D3]([#0,#6,#7,#8])(=O)",
    "L3": "[O;D2]-;!@[#0,#6,#1]",
    "L4": "[C;!D1;!$(C=*)]-;!@[#6]",
    # 'L5':'[N;!D1;!$(N*!-*);!$(N=*);!$(N-[!C;!#0])]-[#0,C]',
    "L5": "[N;!D1;!$(N=*);!$(N-[!#6;!#16;!#0;!#1]);!$([N;R]@[C;R]=O)]",
    "L6": "[C;D3;!R](=O)-;!@[#0,#6,#7,#8]",
    "L7": "[C;D2,D3]-[#6]",
    "#L8": "[C;!R;!D1]-;!@[#6]",
    "L8": "[C;!R;!D1;!$(C!-*)]",
    "L9": "[n;+0;$(n(:[c,n,o,s]):[c,n,o,s])]",
    "L10": "[N;R;$(N(@C(=O))@[C,N,O,S])]",
    "L11": "[S;D2](-;!@[#0,#6])",
    "L12": "[S;D4]([#6,#0])(=O)(=O)",
    "L13": "[C;$(C(-;@[C,N,O,S])-;@[N,O,S])]",
    "L14": "[c;$(c(:[c,n,o,s]):[n,o,s])]",
    "L15": "[C;$(C(-;@C)-;@C)]",
    "L16": "[c;$(c(:c):c)]",
}

BRICS_RETRO_ENVIRONS = {
    "L1": "[C;!D4;!D3](=O)",
    "L3": "[O;D1]-;!@[#0,#6,#1]",
    "L4": "[C;!D4;!$(C=*)]-;!@[#6]",
    # 'L5':'[N;!D1;!$(N*!-*);!$(N=*);!$(N-[!C;!#0])]-[#0,C]',
    "L5": "[N;!D3;!X4;!$(N=*);!$(N-[!#6;!#16;!#0;!#1]);!$([N;R]@[C;R]=O)]",
    "L6": "[C;!D3;!D4;!R](=O)-;!@[#0,#6,#7,#8]",
    "L7": "[C;D1,D2]-[#6]",
    "#L8": "[C;!R;!D4]-;!@[#6]",
    "L8": "[C;!R;!D4;!$(C!-*)]",
    # I updated this one to require an implicit hydrogen
    "L9": "[n;+0;h;$(n(:[c,n,o,s]):[c,n,o,s])]",
    "L10": "[N;D2;R;$(N(@C(=O))@[C,N,O,S])]",
    "L11": "[S;D1](-;!@[#0,#6])",
    "L12": "[S;D3]([#6,#0])(=O)(=O)",
    "L13": "[C;h;$(C(-;@[C,N,O,S,F])-;@[N,O,S])]",
    "L14": "[c;h;$(c(:[c,n,o,s,F]):[n,o,s])]",
    "L15": "[C;h;$(C(-;@C)-;@C)]",
    "L16": "[c;h;$(c(:c):c)]",
}

REACTIONSDEFS = (
    # L1
    [
        ("1", "3", "-"),
        ("1", "5", "-"),
        ("1", "10", "-"),
    ],
    # L3
    [
        ("3", "4", "-"),
        ("3", "13", "-"),
        ("3", "14", "-"),
        ("3", "15", "-"),
        ("3", "16", "-"),
    ],
    # L4
    [
        ("4", "5", "-"),
        ("4", "11", "-"),
    ],
    # L5
    [
        ("5", "12", "-"),
        ("5", "14", "-"),
        ("5", "16", "-"),
        ("5", "13", "-"),
        ("5", "15", "-"),
    ],
    # L6
    [
        ("6", "13", "-"),
        ("6", "14", "-"),
        ("6", "15", "-"),
        ("6", "16", "-"),
    ],
    # L7
    [
        ("7", "7", "="),
    ],
    # L8
    [
        ("8", "9", "-"),
        ("8", "10", "-"),
        ("8", "13", "-"),
        ("8", "14", "-"),
        ("8", "15", "-"),
        ("8", "16", "-"),
    ],
    # L9
    [
        ("9", "13", "-"),
        ("9", "14", "-"),
        ("9", "15", "-"),
        ("9", "16", "-"),
    ],
    # L10
    [
        ("10", "13", "-"),
        ("10", "14", "-"),
        ("10", "15", "-"),
        ("10", "16", "-"),
    ],
    # L11
    [
        ("11", "13", "-"),
        ("11", "14", "-"),
        ("11", "15", "-"),
        ("11", "16", "-"),
    ],
    # L13
    [
        ("13", "14", "-"),
        ("13", "15", "-"),
        ("13", "16", "-"),
    ],
    # L14
    [
        ("14", "14", "-"),
        ("14", "15", "-"),
        ("14", "16", "-"),
    ],
    # L15
    [
        ("15", "16", "-"),
    ],
    # L16
    [
        ("16", "16", "-"),
    ],
)


@lru_cache(maxsize=None)
def getbrics_list():
    all_brics = []
    all_brics_retro = []
    all_brics_type = []
    smartsGps = copy.deepcopy(REACTIONSDEFS)
    for gp in smartsGps:
        for defn in gp:
            g1, g2, bnd = defn
            r1 = BRICS_ENVIRONS["L" + g1]
            r2 = BRICS_ENVIRONS["L" + g2]
            rr1 = BRICS_RETRO_ENVIRONS["L" + g1]
            rr2 = BRICS_RETRO_ENVIRONS["L" + g2]
            g1 = re.sub("[a-z,A-Z]", "", g1)
            g2 = re.sub("[a-z,A-Z]", "", g2)
            # sma = '[$(%s):1].[$(%s):2]>>[$(%s):1]%s[$(%s):2]' % (r1, r2, r1, bnd, r2)
            # should be equivalent to above ...
            retrosma = "[$(%s):1].[$(%s):2]>>[*:1]%s[*:2]" % (rr1, rr2, bnd)
            sma = "[$(%s):1]%s;!@[$(%s):2]>>[%s*]-[*:1].[%s*]-[*:2]" % (
                r1,
                bnd,
                r2,
                g1,
                g2,
            )
            try:
                rx = rdChemReactions.ReactionFromSmarts(sma)
                retro_rx = rdChemReactions.ReactionFromSmarts(retrosma)
                rx.Initialize()
                retro_rx.Initialize()
                all_brics_type.append("L{}-L{}".format(g1, g2))
                all_brics.append(rx)
                all_brics_retro.append(retro_rx)
            except Exception:
                pass
    return all_brics, all_brics_retro, all_brics_type


@lru_cache(maxsize=None)
def get_reactions_list():
    all_rxns = []
    all_rxns_retro = []
    all_rxns_type = []
    with pkg_resources.resource_stream("datamol", "data/reactions.json") as IN:
        rxns = json.load(IN)
        for k, data in rxns.items():
            try:
                rxn = data.get("syn_smarts")
                retro_rxn = data.get("retro_smarts")
                retro_rxn = rdChemReactions.ReactionFromSmarts(retro_rxn)
                rxn = rdChemReactions.ReactionFromSmarts(rxn)
                if (rxn.GetNumReactantTemplates() == 2 and rxn.GetNumProductTemplates() == 1) and (
                    retro_rxn.GetNumReactantTemplates() == 1
                    and retro_rxn.GetNumProductTemplates() == 2
                ):
                    all_rxns.append(retro_rxn)
                    all_rxns_retro.append(rxn)
                    all_rxns_type.append(k)

            except Exception:
                pass
    return all_rxns, all_rxns_retro, all_rxns_type


ALL_RXNS, ALL_RXNS_RETRO, ALL_RXNS_TYPE = get_reactions_list()
ALL_BRICS, ALL_BRICS_RETRO, ALL_BRICS_TYPE = getbrics_list()


def _can_continue_with(mol, rxns):
    for rxn in rxns:
        if rxn.IsMoleculeReactant(mol):
            return True
    return False


def _run_at_all_rct(rxn, mol1, mol2):
    library = []
    rxn = rdChemReactions.ReactionFromSmarts(rdChemReactions.ReactionToSmarts(rxn))
    # display(rxn)
    m1 = rxn.GetReactantTemplate(0)
    m2 = rxn.GetReactantTemplate(1)
    mol1_valid = mol1 is not None
    mol2_valid = mol2 is not None
    isR1 = mol1_valid and mol1.HasSubstructMatch(m1)
    isR2 = mol1_valid and mol1.HasSubstructMatch(m2)
    if isR1 and mol2_valid and mol2.HasSubstructMatch(m2):
        library.extend(rxn.RunReactants((mol1, mol2)))
    if isR2 and mol2_valid and mol2.HasSubstructMatch(m1):
        library.extend(rxn.RunReactants((mol2, mol1)))
    if library:
        library = list(itertools.chain(*library))
    for m in library:
        mol = None
        mSmi = ""
        try:
            mSmi = dm.to_smiles(m)
            mol = dm.to_mol(mSmi)
        except:
            pass
        if mol is None:
            try:
                mol.UpdatePropertyCache()
                mol = dm.sanitize_mol(mol)
                mSmi = dm.to_smiles(m)
                mol = dm.to_mol(mSmi)
            except:
                pass
        if mSmi:
            yield mol, mSmi


def break_mol(
    mol: Chem.rdchem.Mol,
    minFragmentSize: int = 1,
    silent: bool = True,
    onlyUseReactions: list = [],
    randomize: bool = False,
    mode: str = "brics",
    returnTree: bool = False,
):
    """Breaks a molecules into a list of fragment."""

    if mode.lower() == "brics":
        all_reactions = ALL_BRICS
        all_reactions_type = ALL_BRICS_TYPE
    elif mode.lower() == "rxn":
        all_reactions = ALL_RXNS
        all_reactions_type = ALL_RXNS_TYPE
    else:
        all_reactions = ALL_BRICS + ALL_RXNS
        all_reactions_type = ALL_BRICS_TYPE + ALL_RXNS_TYPE
    if randomize:
        p = np.random.permutation(len(all_reactions))
        all_reactions = [all_reactions[ind] for ind in p]
        all_reactions_type = [all_reactions_type[ind] for ind in p]

    nx = dm.graph._get_networkx()
    mSmi = dm.to_smiles(mol, isomeric=True)
    G = nx.DiGraph()
    node_num = 0
    G.add_node(node_num, smiles=mSmi, mol=mol)
    allNodes = set()
    activePool = {mSmi: node_num}
    allNodes.add(mSmi)
    while activePool:
        nSmi = list(activePool.keys())[0]
        parent = activePool.pop(nSmi)
        node = G.nodes[parent]
        mol = node["mol"]
        for rxnIdx, reaction in zip(all_reactions_type, all_reactions):
            if onlyUseReactions and rxnIdx not in onlyUseReactions:
                continue
            ps = reaction.RunReactants((mol,))
            if ps:
                all_pass = [
                    all([prod.GetNumAtoms(onlyExplicit=True) > minFragmentSize for prod in p_])
                    for p_ in ps
                ]
                nz_i = 0
                while nz_i < len(all_pass) and not all_pass[nz_i]:
                    nz_i += 1
                if not silent:
                    print(nSmi, "->", len(ps), "products and selected ", nz_i)
                    # display(MolsToGridImage(list(itertools.chain(*list(ps))), molsPerRow=2))
                prodSeq = ps[nz_i % len(all_pass)]
                seqOk = True
                # we want to disqualify small fragments, so sort the product sequence by size
                prodSeq = [(prod.GetNumAtoms(onlyExplicit=True), prod) for prod in prodSeq]
                prodSeq.sort(key=lambda x: x[0])
                for _, prod in prodSeq:
                    prod.sanitized = True
                    try:
                        Chem.SanitizeMol(prod)
                    except:
                        if dm.sanitize_mol(prod) is None:
                            seqOk = False
                            break
                        continue
                    pSmi = dm.to_smiles(prod, isomeric=True)
                    seqOk = seqOk and (dm.to_mol(pSmi) is not None)

                    notDummies = sum([atm.GetSymbol() != "*" for atm in prod.GetAtoms()])
                    # nDummies = pSmi.count('*')
                    # if minFragmentSize > 0 and (nats - nDummies < minFragmentSize):
                    if minFragmentSize > 0 and notDummies < minFragmentSize:
                        seqOk = False
                        break
                    prod.pSmi = pSmi

                if seqOk:
                    for _, prod in prodSeq:
                        if not prod.sanitized:
                            continue
                        pSmi = prod.pSmi
                        node_num += 1
                        usmi = dm.to_smiles(dm.fix_mol(prod), isomeric=True)
                        G.add_node(node_num, smiles=usmi, mol=prod)
                        G.add_edge(parent, node_num)
                        if usmi not in allNodes:
                            activePool[pSmi] = node_num
                            allNodes.add(usmi)
                    G.nodes[parent]["rxn"] = rxnIdx
                    break  # at least one reaction matches

    leaves_smiles = [
        G.nodes[n]["smiles"] for n in G.nodes() if G.in_degree(n) != 0 and G.out_degree(n) == 0
    ]
    if returnTree:
        return leaves_smiles, allNodes, G
    return leaves_smiles, allNodes


def build(ll_mols, max_n_mols=float("inf"), mode="brics", frag_rxn=None, ADD_RNXS=[]):
    """Build a super molecule from a list of fragments"""

    seen = set()
    stop = False
    CUR_RXNS = []
    CUR_RXNS_TYPE = []

    if mode == "brics":
        CUR_RXNS = ALL_BRICS_RETRO
        CUR_RXNS_TYPE = ALL_BRICS_TYPE
    elif mode == "rxn":
        CUR_RXNS = ALL_RXNS_RETRO
        CUR_RXNS_TYPE = ALL_RXNS_TYPE
    elif mode is not None:
        CUR_RXNS = ALL_BRICS_RETRO + ALL_RXNS_RETRO
        CUR_RXNS_TYPE = ALL_BRICS_TYPE + ALL_RXNS_TYPE

    if ADD_RNXS is not None:
        ADD_RNXS_TYPE = [f"RXN-{i}" for i in range(len(ADD_RNXS))]
        if isinstance(ADD_RNXS, dict):
            ADD_RNXS_TYPE = ADD_RNXS.keys()
            ADD_RNXS = ADD_RNXS.values()
        CUR_RXNS += list(ADD_RNXS)
        CUR_RXNS_TYPE += list(ADD_RNXS_TYPE)

    for i, rxn_type in enumerate(CUR_RXNS_TYPE):
        if (frag_rxn is not None) and (frag_rxn.strip('"') == rxn_type):
            CUR_RXNS = [CUR_RXNS[i]]
            break

    for fraglist in itertools.product(*ll_mols):
        if stop:
            break

        fraglist = list(fraglist)
        for rxn in CUR_RXNS:  # should be size==1 if frag_rxn is provided
            ps = []
            try:
                ps = _run_at_all_rct(rxn, fraglist[0], fraglist[1])
            except Exception:
                pass
            for m, mSmi in ps:
                if len(seen) >= max_n_mols:
                    stop = True
                    break
                if mSmi not in seen:
                    seen.add(mSmi)
                    yield m


def assemble_fragment_order(
    fragmentlist: list,
    seen: Optional[Mol] = None,
    allow_incomplete: bool = False,
    max_n_mols: float = float("inf"),
    RXNS=None,
):
    """Assemble a list of fragment into a set of possible molecules under rules defined by the brics algorithm

    We are of course assuming:

        1. that the order in the fragmentlist matter :D !
        2. that none of the fragment has explicitly defined hydrogen atoms.
        3. only a list of unique molecule is internally maintained

    Args:
        fragmentlist: list of original fragments to grow
        seen: original molecules used as base. If none, the first element of fragment list will be poped out
        allow_incomplete: Whether to accept assembled molecules with missing fragment
    """

    if RXNS is None:
        RXNS = ALL_BRICS_RETRO

    fragmentlist = list(fragmentlist)
    yield_counter = 0
    if seen is None:
        seen = fragmentlist.pop(0)
    seen = [dm.to_smiles(seen)]  # only one molecule to assemble
    while yield_counter < max_n_mols and len(fragmentlist) > 0:
        # find all the way to add this fragment to seen
        frag = fragmentlist.pop(0)
        level_set = [dm.to_mol(x) for x in seen]
        seen = set()
        for sm in level_set:
            try:
                # there is no point in even trying something on molecules that cannot be kekulized
                for rxn in RXNS:
                    for m, mSmi in _run_at_all_rct(rxn, frag, sm):
                        if allow_incomplete and mSmi not in seen:
                            yield m
                            yield_counter += 1
                        seen.add(mSmi)
            except Exception as e:
                print(e)
                pass

    for m in seen:
        if yield_counter < max_n_mols:
            yield dm.to_mol(m)
            yield_counter += 1


def assemble_fragment_iter(
    fragmentlist,
    seens=None,
    scrambleReagents=False,
    max_n_mols=float("inf"),
    maxdepth=3,
    as_smiles=True,
    RXNS=None,
):
    """Perform an assembly from fragment given all potential RXNS transformation."""

    if RXNS is None:
        RXNS = ALL_BRICS_RETRO

    seen = set()
    if max_n_mols <= 0:
        return
    if not seens:
        seens = list(fragmentlist)
    if scrambleReagents:
        seens = list(seens)
        random.shuffle(seens, random=random.random)

    for seen in seens:
        nextSteps = []
        for rxn in RXNS:
            for fg in fragmentlist:
                for m, pSmi in _run_at_all_rct(rxn, fg, seen):
                    if pSmi not in seen:
                        seen.add(pSmi)
                        yield m if not as_smiles else pSmi
                    if _can_continue_with(m, rxn):
                        nextSteps.append(m)

        if nextSteps and len(seen) <= max_n_mols and maxdepth > 0:
            for p in assemble_fragment_iter(
                fragmentlist,
                seens=nextSteps,
                scrambleReagents=scrambleReagents,
                max_n_mols=(max_n_mols - len(seen)),
                maxdepth=maxdepth - 1,
            ):
                pSmi = dm.to_smiles(p, True)
                if pSmi not in seen:
                    seen.add(pSmi)
                    yield p if not as_smiles else pSmi
                    if len(seen) >= max_n_mols:
                        return



from typing import cast

import pytest

import itertools

from rdkit import Chem

import datamol as dm
import numpy as np


def test_to_mol():
    smiles = "O=C(C)Oc1ccccc1C(=O)O"
    mol = dm.to_mol(smiles)
    assert mol.GetNumAtoms() == 13

    smiles = "O=C(C)Oc1ccccc1C(=O)O"
    mol = dm.to_mol(smiles, add_hs=True)
    assert mol.GetNumAtoms() == 21

    smiles = "fake_smiles"
    mol = dm.to_mol(smiles)
    assert mol is None


def test_reorder_atoms():
    smiles = "c1ccc(C(=O)O)c(c1)OC(=O)C"
    mol = dm.to_mol(smiles, add_hs=False, explicit_only=False)

    orders = [a.GetAtomicNum() for a in mol.GetAtoms()]
    assert orders == [6, 6, 6, 6, 6, 8, 8, 6, 6, 8, 6, 8, 6]

    mol = dm.reorder_atoms(mol)
    orders = [a.GetAtomicNum() for a in mol.GetAtoms()]
    assert orders == [6, 8, 8, 8, 6, 6, 6, 6, 8, 6, 6, 6, 6]


def test_randomize_atoms():
    smiles = "c1ccc(C(=O)O)c(c1)OC(=O)C"
    mol = dm.to_mol(smiles)
    orders = [a.GetAtomicNum() for a in mol.GetAtoms()]

    randomized_mol = dm.randomize_atoms(mol)
    randomized_orders = [a.GetAtomicNum() for a in randomized_mol.GetAtoms()]

    assert sum(orders) == sum(randomized_orders)


def test_to_neutral():
    smiles = "[NH4+]"
    mol = dm.to_mol(smiles, add_hs=False, explicit_only=False)

    smiles = dm.to_smiles(dm.to_neutral(mol))
    assert smiles == "[NH4]"

    smiles = "O=C(c1ccccc1)[O-]"
    mol = dm.to_mol(smiles, add_hs=False, explicit_only=False)
    uncharged_mol = dm.to_neutral(mol)
    assert sum([a.GetFormalCharge() for a in uncharged_mol.GetAtoms()]) == 0


def test_sanitize():
    smiles = "CC(=O)Oc1ccccc1C(=O)O"
    mol = dm.to_mol(smiles, sanitize=False)
    mol = dm.sanitize_mol(mol, charge_neutral=True)
    assert dm.to_smiles(mol) == "CC(=O)Oc1ccccc1C(=O)O"

    mol = dm.sanitize_mol(None, charge_neutral=True)
    assert mol is None

    smiles_list = (
        "CC.[H][N:1]1(C)=CC(O)=CC2CCCCC12",  # broken
        "O=c1ccc2ccccc2n1",  # sanitize
        "Cc1nnnn1C",  # none
        "CCc1ccc2nc(=O)c(cc2c1)Cc1nnnn1C1CCCCC1",  # sanitize
        "c1cnc2cc3ccnc3cc12",  # none
        "c1cc2cc3ccnc3cc2n1",  # none
        "O=c1ccnc(c1)-c1cnc2cc3ccnc3cc12",  # sanitize
        "O=c1ccnc(c1)-c1cc1",  # broken
    )

    # check sanitize_mol
    assert dm.to_mol(smiles_list[1]) is None
    assert dm.to_mol(smiles_list[2]) is not None
    assert dm.sanitize_mol(None) is None
    assert dm.sanitize_mol(dm.to_mol(smiles_list[0], sanitize=False)) is None
    assert dm.sanitize_mol(dm.to_mol(smiles_list[1], sanitize=False)) is not None
    assert dm.sanitize_mol(dm.to_mol(smiles_list[2], sanitize=False)) is not None

    mol_2 = dm.sanitize_mol(dm.to_mol(smiles_list[1], sanitize=False))
    assert dm.to_smiles(mol_2) == dm.sanitize_smiles("O=c1ccc2ccccc2[nH]1")

    fixed_smiles = [dm.sanitize_smiles(smiles) for smiles in smiles_list]
    assert len([x for x in fixed_smiles if x is not None]) == 6

    smiles = "CCCN(=O)=O"
    mol = dm.to_mol(smiles, sanitize=False)
    sane_mol = dm.sanitize_mol(mol)
    assert dm.to_smiles(sane_mol) == "CCC[N+](=O)[O-]"


def test_sanitize_first():
    smiles = ["fake_smiles", "CC(=O)Oc1ccccc1C(=O)O"]
    mols = [dm.to_mol(s) for s in smiles]
    mol = dm.sanitize_first(mols)
    assert dm.to_smiles(mol) == "CC(=O)Oc1ccccc1C(=O)O"


def test_standardize_mol():
    sm = "[Na]OC1=CC2CCCCC2N=C1"
    sm_standard = dm.to_smiles(dm.standardize_smiles(sm))
    standard_mol = dm.standardize_mol(dm.to_mol(sm), disconnect_metals=True, uncharge=True)
    mol_standard = dm.to_smiles(Chem.MolToSmiles(standard_mol))
    assert sm_standard == mol_standard


def test_fix_valence():
    sm = "Cl.[H][N:1]1=CC(O)=CC2CCCCC12"
    mol = Chem.MolFromSmiles(sm, sanitize=False)
    mol.UpdatePropertyCache(False)
    mol_copy = dm.copy_mol(mol)

    nitrogen_atom = [a for a in mol.GetAtoms() if a.GetAtomMapNum() == 1][0]
    nitrogen_valence = nitrogen_atom.GetExplicitValence()
    assert dm.incorrect_valence(nitrogen_atom, True)

    fixed_mol = dm.fix_valence_charge(mol, inplace=False)
    assert dm.to_mol(Chem.MolToSmiles(fixed_mol)) is not None

    # expect nitrogen atom to still be incorrect
    assert dm.incorrect_valence(nitrogen_atom, True)

    # in place fix
    fixed_mol = dm.fix_valence_charge(mol, inplace=True)
    # nitrogen should be charged positively if this was fixed.
    assert nitrogen_atom.GetFormalCharge() == 1

    fixed_mol2 = dm.fix_valence(mol_copy)
    fixed_nitrogen_atom = [a for a in fixed_mol2.GetAtoms() if a.GetAtomMapNum() == 1][0]
    assert fixed_nitrogen_atom.GetExplicitValence() < nitrogen_valence

    # mol should be fixed
    assert dm.to_mol(Chem.MolToSmiles(fixed_mol2)) is not None


def test_adjust_singleton():
    sm = "Cl.[N:1]1=CC(O)=CC2CCCCC12.CC.C"
    mol = dm.to_mol(sm)
    fixed_mol = dm.adjust_singleton(mol)
    assert len(Chem.rdmolops.GetMolFrags(fixed_mol)) == 2
    assert fixed_mol.HasSubstructMatch(Chem.MolFromSmiles("CC"))  # assert ethyl is there


def test_fixmol():
    sm = "C.Cl.CC.[H][N:1]1(C)=CC(O)=CC2CCCCC12"
    mol = Chem.MolFromSmiles(sm, sanitize=False)
    # mol.UpdatePropertyCache(False)
    # Chem.Kekulize(mol)
    res = dm.fix_mol(mol, n_iter=1)  # copy by default

    # should still be invalid in term of valence for nitrogen
    assert not dm.incorrect_valence(res)

    res2 = dm.fix_mol(mol, n_iter=2)
    # not expecting difference between res2 and res3
    assert Chem.MolToSmiles(res) == Chem.MolToSmiles(res2)

    # only largest expected_here
    res_largest = dm.fix_mol(mol, largest_only=True)

    dm.fix_mol(mol, remove_singleton=True, largest_only=True)
    assert len(Chem.rdmolops.GetMolFrags(res_largest)) == 1

    expected_largest_fix = dm.standardize_smiles("OC1=CC2CCCCC2[N:1]=C1")
    assert dm.standardize_smiles(Chem.MolToSmiles(res_largest)) == expected_largest_fix

    res_no_singleton = dm.fix_mol(mol, n_iter=2, remove_singleton=True)
    assert len(Chem.rdmolops.GetMolFrags(res_largest)) == 1
    assert len(Chem.rdmolops.GetMolFrags(res_no_singleton)) == 2


def test_dative_bond():
    smis = "CC1=CC=CC(=C1N\\2O[Co]3(ON(\\C=[N]3\\C4=C(C)C=CC=C4C)C5=C(C)C=CC=C5C)[N](=C2)\\C6=C(C)C=CC=C6C)C"
    expected_result = (
        "CC1=CC=CC(C)=C1N1C=N(C2=C(C)C=CC=C2C)->[Co]2(<-N(C3=C(C)C=CC=C3C)=CN(C3=C(C)C=CC=C3C)O2)O1"
    )

    assert dm.is_transition_metal(Chem.Atom("Co"))

    # sodium is not a transition metal
    assert not dm.is_transition_metal(Chem.Atom("Na"))

    mol = dm.set_dative_bonds(Chem.MolFromSmiles(smis, sanitize=False))
    assert Chem.MolToSmiles(mol) == expected_result
    assert dm.to_mol(Chem.MolToSmiles(mol)) is not None


def test_copy_mol():
    mol = dm.to_mol("OC1=CC2CCCCC2[N:1]=C1")
    new_mol = dm.copy_mol(mol)

    assert dm.to_smiles(mol) == dm.to_smiles(new_mol)


def test_set_mol_props():
    mol = dm.to_mol("CCC")

    props = {}
    props["number"] = 55
    props["float"] = 5.555
    props["string"] = "hello"
    props["something_else"] = type(int)

    dm.set_mol_props(mol, props)

    mol_props = mol.GetPropsAsDict()
    assert mol_props["number"] == props["number"]
    assert mol_props["float"] == props["float"]
    assert mol_props["string"] == props["string"]
    assert mol_props["something_else"] == str(props["something_else"])

    dm.set_mol_props(mol, props, copy=True)


def test_set_mol_props_overflow():
    mol = dm.to_mol("CCC")
    dm.set_mol_props(mol, dict(hello=661440088496))
    assert mol.GetPropsAsDict() == {"hello": 661440088496}


def test_copy_mol_props():
    source = dm.to_mol("CCC")
    destination = dm.to_mol("CC")

    props = {}
    props["bool"] = True
    props["number"] = 55
    props["float"] = 5.555
    props["string"] = "hello"
    props["something_else"] = type(int)

    dm.set_mol_props(source, props)

    dm.copy_mol_props(source, destination)

    assert destination.GetPropsAsDict() == source.GetPropsAsDict()


def test_atom_indices_to_mol():
    mol: dm.Mol = dm.to_mol("OC1=CC2CCCCC2[N:1]=C1")

    mol2 = dm.atom_indices_to_mol(mol)
    for atom in mol2.GetAtoms():
        assert atom.GetIntProp("molAtomMapNumber") == atom.GetIdx()

    mol3 = dm.atom_indices_to_mol(mol, copy=True)
    for atom in mol3.GetAtoms():
        assert atom.GetIntProp("molAtomMapNumber") == atom.GetIdx()


def test_to_mol_bad_input():
    with pytest.raises(ValueError):
        dm.to_mol(555)


def test_to_mol_input_mol():
    mol = dm.to_mol(dm.to_mol("CCC"))
    assert dm.to_smiles(mol) == "CCC"


def test_to_mol_ordered():
    s1 = dm.to_smiles(dm.to_mol("C1(O)=CC2CCCCC2[N:1]=C1", ordered=False), canonical=False)
    s2 = dm.to_smiles(dm.to_mol("C1(O)=CC2CCCCC2[N:1]=C1", ordered=True), canonical=False)
    assert s1 != s2


def test_to_mol_kekulize():
    smiles = "C1=CC=CN=C1"
    dm.to_mol(smiles, kekulize=True)


def test_reorder_atoms_without_atoms():
    mol1 = dm.to_mol("")
    mol2 = dm.reorder_atoms(mol1)
    assert dm.to_smiles(mol1) == dm.to_smiles(mol2)


def test_randomize_atoms_without_atoms():
    mol1 = dm.to_mol("")
    mol2 = dm.randomize_atoms(mol1)
    assert dm.to_smiles(mol1) == dm.to_smiles(mol2)


def test_to_neutral_without_atoms():
    assert dm.to_neutral(None) is None


def test_sanitize_smiles_none():
    assert dm.sanitize_smiles(444) is None


def test_standardize_smiles_tautomer():
    smiles = "C1=CC=CN=C1"
    std_smiles = dm.standardize_smiles(smiles, tautomer=True)
    assert "c1ccncc1" == std_smiles


def test_decrease_bond():
    smiles = "C=CCC#C"
    mol = dm.to_mol(smiles)

    single_bond = mol.GetBondBetweenAtoms(1, 2)
    double_bond = mol.GetBondBetweenAtoms(0, 1)
    triple_bond = mol.GetBondBetweenAtoms(3, 4)

    assert dm.decrease_bond(single_bond) is None
    assert dm.decrease_bond(double_bond) == dm.SINGLE_BOND
    assert dm.decrease_bond(triple_bond) == dm.DOUBLE_BOND


def test_replace_dummies_atoms():
    smiles = "C=CCC*"
    mol = dm.to_mol(smiles)

    mol2 = dm.replace_dummies_atoms(mol)
    assert dm.to_smiles(mol2) == "C=CCCC"


def test_keep_largest_frag():
    mol = dm.to_mol("Cl.[N:1]1=CC(O)=CC2CCCCC12.CC.C")
    frag = dm.keep_largest_fragment(mol)
    assert dm.to_smiles(frag) == "OC1=CC2CCCCC2[N:1]=C1"


def test_sanitize_mol_keep_props_and_conformers():
    # Generate a mol with props and a conformer
    props = dict(test_int=588, test_str="hello")
    smiles = "CCC[N+](=O)[O-]"

    mol = dm.to_mol(smiles)
    mol = dm.set_mol_props(mol, props)
    mol = dm.conformers.generate(mol, n_confs=1)
    pos = mol.GetConformer().GetPositions()

    # Sanitize
    sane_mol = dm.sanitize_mol(mol)

    # Check properties
    assert sane_mol.GetPropsAsDict() == props

    # Check conformer
    conf = sane_mol.GetConformer()
    assert sane_mol.GetNumConformers() == 1
    assert conf.Is3D()
    np.testing.assert_almost_equal(conf.GetPositions(), pos, decimal=4)


def test_sanitize_mol_multiple_conformers_warning(caplog):
    # Generate a mol with props and a conformer
    smiles = "CCC[N+](=O)[O-]"

    mol = dm.to_mol(smiles)
    mol = dm.conformers.generate(mol, n_confs=10)

    # Check warning log
    dm.sanitize_mol(mol)
    assert "WARNING" != caplog.text
    assert (
        "The molecule contains multiple conformers. Only the first one will be preserved."
        in caplog.text
    )


def test_sanitize_mol_multiple_conformers_no_warning(caplog):
    # Generate a mol with props and a conformer
    smiles = "CCC[N+](=O)[O-]"

    mol = dm.to_mol(smiles)
    mol = dm.conformers.generate(mol, n_confs=10)

    # Check no warning log
    dm.sanitize_mol(mol, verbose=False)
    assert caplog.text == ""


def test_same_mol():
    mol1 = dm.to_mol("CC(=O)Oc1ccccc1C(=O)O")
    mol2 = dm.to_mol("C1OC1CC")

    assert dm.same_mol(mol1, mol2) is False
    assert dm.same_mol(mol1, mol1) is True
    assert dm.same_mol(mol2, mol2) is True
    assert dm.same_mol(None, mol2) is False
    assert dm.same_mol(mol1, None) is False
    assert dm.same_mol(None, None) is False

    # check with not sane molecule
    mol1 = dm.to_mol("c1ccccc1")
    mol2 = dm.to_mol("C1=CC=CC=C1", sanitize=False)  # not sane benzene but still a benzene molecule

    assert dm.same_mol(mol1, mol2) is True
    assert dm.same_mol(mol1, mol1) is True
    assert dm.same_mol(mol2, mol2) is True
    assert dm.same_mol(None, mol2) is False
    assert dm.same_mol(mol1, None) is False
    assert dm.same_mol(None, None) is False

    mol1 = dm.to_mol("N=C(N)O")
    mol2 = dm.to_mol("NC(N)=O")

    assert dm.same_mol(mol1, mol2) is True
    assert dm.same_mol(mol1, mol2, use_non_standard_inchikey=True) is False


def test_atom_list_to_bond():
    mol = dm.to_mol("CC(=O)OC1=CC=CC=C1C(=O)O")

    bond_indices = dm.atom_list_to_bond(mol, atom_indices=[3, 4, 5, 6, 10, 11], bond_as_idx=True)
    assert bond_indices == [3, 4, 5, 10]

    bonds = dm.atom_list_to_bond(mol, atom_indices=[3, 4, 5, 6, 10, 11], bond_as_idx=False)
    assert [b.GetIdx() for b in bonds] == [3, 4, 5, 10]


def test_protect_atoms():
    mol = dm.to_mol("CC(=O)Oc1cnccc1C(O)=O")

    # carbonyl group
    query = dm.from_smarts("[$([CX3]=[OX1]),$([CX3+]-[OX1-])]")

    # find id of single nitrogen
    atom_ids = [at.GetIdx() for at in mol.GetAtoms() if at.GetSymbol() == "N"]  # type: ignore
    protected_atoms = itertools.chain(*mol.GetSubstructMatches(query))  # type: ignore
    protected_atoms = set(list(protected_atoms) + atom_ids)
    new_mol = dm.protect_atoms(mol, substruct=query, atoms=atom_ids)
    computed_protect_atoms = set(
        [
            at.GetIdx()
            for at in new_mol.GetAtoms()
            if int(at.GetPropsAsDict().get("_protected", 0)) == 1
        ]
    )
    assert protected_atoms == computed_protect_atoms


def test_add_remove_hs():
    smiles = "OC1=CC2CCCCC2[N:1]=C1"

    mol = dm.to_mol(smiles)

    mol2 = dm.add_hs(mol)
    assert (
        dm.to_smiles(mol2)
        == "[H]OC1=C([H])C2([H])C([H])([H])C([H])([H])C([H])([H])C([H])([H])C2([H])[N:1]=C1[H]"
    )

    mol3 = dm.remove_hs(mol)
    assert dm.to_smiles(mol3) == smiles


def test_unique_id():
    smiles_list = [
        # mols coming from FreeSolv
        "c1ccc2c(c1)Cc3ccccc3C2",
        "CCCC/C=C/C",
        "Cc1c[nH]cn1",
        "Cc1ccc(nc1)C",
        "Cc1cccs1",
        "C(CO[N+](=O)[O-])CO[N+](=O)[O-]",
        "C(=C(F)F)(C(F)(F)F)F",
        "CCOP(=S)(OCC)SCSP(=S)(OCC)OCC",
        "CC1=CC(=O)[C@@H](CC1)C(C)C",
        "Cc1cc(ccc1Cl)O",
        "CC(C)O",
        "C",
        "CCCC(=O)C",
        "CCCCCCCC=O",
        "CCc1cccc2c1cccc2",
        # Two cherry picked tautomeric forms
        # of the same molecule.
        "N=C(N)O",
        "NC(N)=O",
        # an invalid molecule
        "C(C)(C)(C)(C)(C)",
        # a molecule with a dummy atom
        "CCCC*",
    ]

    mols = [dm.to_mol(s) for s in smiles_list]
    ids = [dm.unique_id(m) for m in mols]
    assert ids == [
        "2f0a84f817a369e79a2b44085ebe3a27",
        "3dc94b495eba316c2df4563457c77842",
        "df2c2dcbe7262f02f371d5523248e3d6",
        "1d0bc775e2b6ab0a077548a26ce9586f",
        "6cb8a819d03a38a6ed4771d1bbfd899c",
        "ae2683fabd3b5c31ac0554c213e1f37a",
        "c77fd251ff9b35379e8b08f1cf870391",
        "1a239153353f30ec3067af0461a5bff3",
        "f5bcdce1e7726f01abdc8f30f627ebcf",
        "4121e41877275378c2ae501815ab2876",
        "a6e46ea133142499b60f73a6d3e94248",
        "cbaaa6452f9cc0649cc9920a698c6f5d",
        "70572a2d32cf4707138a6ef81cfd7253",
        "180d723f0fbcc955a975d740fbc91adf",
        "11e555af02fa8342860752227c0ddc9e",
        # The two IDS of the tautomers are indeed different.
        "3a04e3e47b86287b035b4affe84dff91",
        "943d4548ef05069b79e929ed140ce905",
        # unsupported syntax should return None
        None,
        None,
    ]

    assert dm.unique_id(None) is None


def test_clear_mol_props():
    smiles = "Cc1ccc(CO)cc1-c1ccc2c(n1)n(C)c(=O)n2CC(C)(C)C"
    mol = dm.to_mol(smiles)

    # Set properties to the molecule
    props = dict(myname="hello", a_digit=99)
    mol = dm.set_mol_props(mol, props)

    # Check
    assert "myname" in mol.GetPropsAsDict()
    assert "a_digit" in mol.GetPropsAsDict()

    # Clear all the properties
    mol2 = dm.clear_mol_props(mol)

    # Check
    assert "myname" not in mol2.GetPropsAsDict()
    assert "a_digit" not in mol2.GetPropsAsDict()

    # Clear only a single properties
    mol3 = dm.clear_mol_props(mol, property_keys=["a_digit"])

    # Check
    assert "myname" in mol3.GetPropsAsDict()
    assert "a_digit" not in mol3.GetPropsAsDict()

    # Clear only a single properties (from a string key)
    mol3 = dm.clear_mol_props(mol, property_keys="a_digit")

    # Check
    assert "myname" in mol3.GetPropsAsDict()
    assert "a_digit" not in mol3.GetPropsAsDict()


def test_strip_mol_to_core():
    mol = dm.to_mol("CC(=O)NC1CCC2=CC(=C(C(=C2C3=CC=C(C(=O)C=C13)OC)OC)OC)OC")
    mol2 = dm.strip_mol_to_core(mol)

    assert dm.to_inchikey(mol2) == "XSMDDAFPLXKNOA-UHFFFAOYSA-N"


def test_make_scaffold_generic():
    # NOTE(hadim): pretty sure doing assert on SMARTS string is fragile and might change
    # in the future RDKit versions. So... hold and wait for it to break xD

    mol = dm.to_mol("CC(=O)NC1CCC2=CC(=C(C(=C2C3=CC=C(C(=O)C=C13)OC)OC)OC)OC")
    mol2 = dm.make_scaffold_generic(mol)
    assert (
        dm.to_smarts(mol2)
        == "[#0]-[#0](=[#0])-[#0]-[#0]1-[#0]-[#0]-[#0]2:[#0]:[#0](:[#0](:[#0](:[#0]:2-[#0]2:[#0]:[#0]:[#0](:[#0](=[#0]):[#0]:[#0]:2-1)-[#0]-[#0])-[#0]-[#0])-[#0]-[#0])-[#0]-[#0]"
    )

    mol = dm.to_mol("CC(=O)NC1CCC2=CC(=C(C(=C2C3=CC=C(C(=O)C=C13)OC)OC)OC)OC")
    mol2 = dm.make_scaffold_generic(mol, include_bonds=True)
    assert (
        dm.to_smarts(mol2)
        == "[#0][#0]([#0])[#0][#0]1[#0][#0][#0]2:[#0]:[#0](:[#0](:[#0](:[#0]:2[#0]2:[#0]:[#0]:[#0](:[#0]([#0]):[#0]:[#0]1:2)[#0][#0])[#0][#0])[#0][#0])[#0][#0]"
    )


def test_to_scaffold():
    mol = dm.to_mol("CC(=O)NC1CCC2=CC(=C(C(=C2C3=CC=C(C(=O)C=C13)OC)OC)OC)OC")
    mol2 = dm.to_scaffold_murcko(mol)
    assert dm.to_inchikey(mol2) == "XSMDDAFPLXKNOA-UHFFFAOYSA-N"

    mol = dm.to_mol("CC(=O)NC1CCC2=CC(=C(C(=C2C3=CC=C(C(=O)C=C13)OC)OC)OC)OC")
    mol2 = dm.to_scaffold_murcko(mol, make_generic=True)
    assert (
        dm.to_smarts(mol2)
        == "[#0]1-[#0]-[#0]-[#0]2:[#0]:[#0]:[#0]:[#0]:[#0]:2-[#0]2:[#0]:[#0]:[#0]:[#0](=[#0]):[#0]:[#0]:2-1"
    )


def test_compute_ring_systems():
    mol = dm.to_mol("CC(=O)NC1CCC2=CC(=C(C(=C2C3=CC=C(C(=O)C=C13)OC)OC)OC)OC")

    systems = dm.compute_ring_system(mol)
    assert len(systems) == 1
    assert len(systems[0]) == 16
    assert isinstance(systems, list)
    assert isinstance(systems[0], set)

    mol = dm.to_mol("CN1C(=O)CN=C(C2=C1C=CC(=C2)Cl)C3=CC=CC=C3")

    systems = dm.compute_ring_system(mol)
    assert len(systems) == 2
    assert len(systems[0]) == 11
    assert len(systems[1]) == 6
    assert isinstance(systems, list)
    assert isinstance(systems[0], set)


def test_hash_mol():
    # rdkit.Chem.RegistrationHash is not available for rdkit before
    # 2022.09.
    if dm.is_lower_than_current_rdkit_version("2022.09"):
        with pytest.raises(NotImplementedError):
            dm.hash_mol(dm.to_mol("CCC"))

        return

    with pytest.raises(ValueError):
        dm.hash_mol(dm.to_mol("CCC"), hash_scheme="invalid")

    data = [
        {
            "smiles": "Clc1ccc2ccccc2c1",
            "all": "334adc7dd6bf2b9384e1d5dee8a983e57bbaed80",
            "no_stereo": "6c0b4720e29e47be5062b0f13d71443a6fead14a",
            "no_tautomers": "b92bca205301bf115e5acbac38827f2fc3f3b572",
        },
        {
            "smiles": "CCCCCCCCCCCCO",
            "all": "4827ecd92b1b8b865d7f9e9c47fe8aadc5827a72",
            "no_stereo": "b2ae43fb4061ec507ffb11a61b009037500ac010",
            "no_tautomers": "31cf7325e8beb6c46dfee7ce9968252309317b1b",
        },
        {
            "smiles": "Nc1cnn(-c2ccccc2)c(=O)c1Br",
            "all": "5a2bb241fb005ba2403bbd8c25453fb714fca3b0",
            "no_stereo": "0b72a56d0b499e4e9bf58975b74ecaa86e50373a",
            "no_tautomers": "32c8840b71755d5a859963be70e8db6bd29c01e7",
        },
        {
            "smiles": "Cn1c(=O)[nH]c2[nH]c(=O)[nH]c2c1=O",
            "all": "2a1f545077bec17e3b2a6f6819dda0e55eab4442",
            "no_stereo": "8d7944f78fb61a6199f90f1505c6249b317272c0",
            "no_tautomers": "ed2f9eb1f7ce5a6ba28fad24074f975470d6937c",
        },
        {
            "smiles": "CN1CCCC(CN2c3ccccc3Sc3ccccc32)C1",
            "all": "419aef09d075056837b15da108b87f6737436de2",
            "no_stereo": "389d7084f68935d52f51fe850f72a2294e04acb3",
            "no_tautomers": "6e9142b90e22152f73c2476a5531632a32f7b8c7",
        },
        {
            "smiles": "Cc1cc(OC(=O)N(C)C)n(-c2ccccc2)n1",
            "all": "f01015c60174e7efe1fee7eb6b64804da739d104",
            "no_stereo": "274054a2de68853740504b260f6dec88cffaac28",
            "no_tautomers": "68820d4a5e08bd5476b99684192fbd73d7da64bf",
        },
        {
            "smiles": "O=[N+]([O-])c1ccc(Oc2ccc(Cl)cc2Cl)cc1",
            "all": "cfff22a4055796418862d0cdf86126118113edd0",
            "no_stereo": "f077a2f705f7e00c9206230c9c78df6641673dd8",
            "no_tautomers": "3df1b08580a5615e305cb8c5eb86f519dad6257f",
        },
        {
            "smiles": "Cc1ncc([N+](=O)[O-])n1CCO",
            "all": "973825b3a38bf44b7f264daad8c4d7631c123ec4",
            "no_stereo": "7678c0d9bb81169afd4ddf6d6f0503b126c9e1d7",
            "no_tautomers": "407c5df463de39b4605132d5f28273637952d000",
        },
        {
            "smiles": "Cc1cc(C)c2ccccc2c1",
            "all": "f2e834775b5814cf7b6d503677b5751821ea2f5c",
            "no_stereo": "1fe225caa22af11e2c32320fc6d11a8819168b92",
            "no_tautomers": "5e566a9bab54c6e733e432001428e2f0d36767cc",
        },
        {
            "smiles": "Oc1cccc2cccnc12",
            "all": "2645e016677cf9e82977b8a08a3d207e7e571f51",
            "no_stereo": "0bdcdca39192466f42f547dcd084772d13fd5f0e",
            "no_tautomers": "2f4c569ee7c73804a11758d6d16aa411e7531571",
        },
    ]

    for datum in data:
        mol = dm.to_mol(datum["smiles"])

        hash_value = dm.hash_mol(mol, hash_scheme="all")
        assert hash_value == datum["all"]

        hash_value = dm.hash_mol(mol, hash_scheme="no_stereo")
        assert hash_value == datum["no_stereo"]

        hash_value = dm.hash_mol(mol, hash_scheme="no_tautomers")
        assert hash_value == datum["no_tautomers"]


def test_to_mol_keep_hs():
    smiles = "[H]OC([H])([H])c1c([H])c([H])c(C([H])([H])[H])c(-c2nc3c(c([H])c2[H])n(C([H])([H])C(C([H])([H])[H])(C([H])([H])[H])C([H])([H])[H])c(=O)n3C([H])([H])[H])c1[H]"

    mol = dm.to_mol(smiles)
    mol = cast(dm.Mol, mol)
    assert len(mol.GetAtoms()) == 25

    mol = dm.to_mol(smiles, remove_hs=False)
    mol = cast(dm.Mol, mol)
    assert len(mol.GetAtoms()) == 50


def test_clear_atom_props():
    smiles = "Cc1ccc(CO)cc1-c1ccc2c(n1)n(C)c(=O)n2CC(C)(C)C"
    mol = dm.to_mol(smiles)

    # add the `molAtomMapNumber` property to the atoms
    mol = dm.atom_indices_to_mol(mol)

    # Check
    assert all(["molAtomMapNumber" in a.GetPropsAsDict() for a in mol.GetAtoms()])

    # Remove all the properties
    mol2 = dm.clear_atom_props(mol)

    # Check
    assert not all(["molAtomMapNumber" in a.GetPropsAsDict() for a in mol2.GetAtoms()])

    # Remove only a single property
    mol2 = dm.clear_atom_props(mol, property_keys=["molAtomMapNumber"])

    # Check
    assert not all(["molAtomMapNumber" in a.GetPropsAsDict() for a in mol2.GetAtoms()])

    # Remove only a single property from a string key
    mol3 = dm.clear_atom_props(mol, property_keys="molAtomMapNumber")

    # Check
    assert not all(["molAtomMapNumber" in a.GetPropsAsDict() for a in mol3.GetAtoms()])


def test_clear_atom_map_number():
    smiles = "Cc1ccc(CO)cc1-c1ccc2c(n1)n(C)c(=O)n2CC(C)(C)C"
    mol = dm.to_mol(smiles)

    # add the `molAtomMapNumber` property to the atoms
    mol = dm.atom_indices_to_mol(mol)

    # Check
    assert all(["molAtomMapNumber" in a.GetPropsAsDict() for a in mol.GetAtoms()])

    # Remove all the properties
    mol2 = dm.clear_atom_map_number(mol)

    # Check
    assert not all(["molAtomMapNumber" in a.GetPropsAsDict() for a in mol2.GetAtoms()])


def test_get_atom_positions():
    smiles = "[H:14][c:5]1[c:3]([c:7]([c:4]([c:6]([c:8]1[N:10]([H:18])[C:2](=[N+:11]([H:19])[H:20])[N:9]([H:16])[H:17])[H:15])[H:13])[F:1])[H:12]"
    mol = dm.to_mol(smiles, remove_hs=False)
    mol = dm.conformers.generate(mol, n_confs=1, add_hs=False)

    # Get positions
    positions_1 = dm.get_atom_positions(mol, reorder_to_atom_map_number=False)
    positions_2 = dm.get_atom_positions(mol, reorder_to_atom_map_number=True)

    # both arrays should not be equal
    assert not np.allclose(positions_1, positions_2)

    # but their sums should be
    assert np.allclose(np.sum(positions_1), np.sum(positions_2))


def test_get_atom_positions_fails():
    smiles = "CCCO"
    mol = dm.to_mol(smiles)

    with pytest.raises(ValueError):
        dm.get_atom_positions(mol)

    mol_with_conf = dm.conformers.generate(mol, n_confs=1)

    dm.get_atom_positions(mol_with_conf, reorder_to_atom_map_number=False)

    with pytest.raises(ValueError):
        dm.get_atom_positions(mol_with_conf, reorder_to_atom_map_number=True)


def test_set_atom_positions():
    smiles = "[H:14][c:5]1[c:3]([c:7]([c:4]([c:6]([c:8]1[N:10]([H:18])[C:2](=[N+:11]([H:19])[H:20])[N:9]([H:16])[H:17])[H:15])[H:13])[F:1])[H:12]"

    mol = dm.to_mol(smiles, remove_hs=False)

    positions = [
        [1.7, -6.67, 3.15],
        [0.2, 4.72, 0.78],
        [3.54, -2.64, 2.88],
        [0.43, -3.87, -0.09],
        [3.44, -0.2, 1.8],
        [0.02, -1.5, -1.0],
        [2.12, -4.54, 1.9],
        [1.5, 0.48, 0.02],
        [0.53, 7.24, 0.25],
        [1.17, 2.91, -0.85],
        [-1.22, 4.15, 2.71],
        [4.64, -3.24, 4.55],
        [-0.89, -5.43, -0.78],
        [4.52, 1.43, 2.45],
        [-1.45, -1.02, -2.48],
        [-0.15, 8.68, 1.38],
        [1.65, 7.88, -1.21],
        [2.24, 3.64, -2.15],
        [-1.96, 2.4, 3.0],
        [-2.02, 5.59, 3.71],
    ]
    positions = np.array(positions)

    # Using the atom map numbers
    mol2 = dm.set_atom_positions(
        mol=mol,
        positions=positions,
        conf_id=0,
        use_atom_map_numbers=True,
    )

    # Check the 3d flag is set
    assert mol2.GetConformers()[0].Is3D()

    # Here the ordering has been changed so only the sum will be equal
    conformer = mol2.GetConformers()[0]
    assert np.allclose(conformer.GetPositions().sum(), positions.sum())

    # Without using the atom map numbers
    # Note that in that case, the conformer will be messed up here since
    # the input positions are mapped to the atom map numbers
    mol2 = dm.set_atom_positions(
        mol=mol,
        positions=positions,
        conf_id=0,
        use_atom_map_numbers=False,
    )

    # Here the order has been kept so the positions must match
    conformer = mol2.GetConformers()[0]
    np.allclose(conformer.GetPositions(), positions)

    # Check 3d flag is not set
    positions_2d = positions
    positions_2d[:, 2] = 0

    mol3 = dm.set_atom_positions(
        mol=mol,
        positions=positions,
        conf_id=0,
        use_atom_map_numbers=True,
    )

    assert not mol3.GetConformers()[0].Is3D()


def test_set_atom_positions_fails():
    smiles = "CCCO"
    mol = dm.to_mol(smiles)

    positions = [
        [1.7, -6.67, 3.15],
        [0.2, 4.72, 0.78],
        [3.54, -2.64, 2.88],
        [0.43, -3.87, -0.09],
    ]
    positions = np.array(positions)

    # Check it works
    dm.set_atom_positions(
        mol=mol,
        positions=positions,
        conf_id=0,
        use_atom_map_numbers=False,
    )

    # Use atom map numbers but the prop is not set.
    with pytest.raises(ValueError):
        dm.set_atom_positions(
            mol=mol,
            positions=positions,
            conf_id=0,
            use_atom_map_numbers=True,
        )

    # Wrong number of dimensions of the positions array
    with pytest.raises(ValueError):
        dm.set_atom_positions(
            mol=mol,
            positions=[positions],
            conf_id=0,
            use_atom_map_numbers=True,
        )

    # Wrong shape of `positions`
    with pytest.raises(ValueError):
        dm.set_atom_positions(
            mol=mol,
            positions=positions[1:, :],
            conf_id=0,
            use_atom_map_numbers=True,
        )



try:
    from importlib.metadata import version
    from importlib.metadata import PackageNotFoundError
except ModuleNotFoundError:
    # Try backported to PY<38 `importlib_metadata`.
    from importlib_metadata import version
    from importlib_metadata import PackageNotFoundError


import rdkit
import packaging.version


try:
    __version__ = version("datamol")
except PackageNotFoundError:
    # package is not installed
    __version__ = "dev"

CURRENT_RDKIT_VERSION = rdkit.__version__
CURRENT_RDKIT_VERSION_OBJ = packaging.version.parse(CURRENT_RDKIT_VERSION)


def is_lower_than_current_rdkit_version(rdkit_version: str):
    return CURRENT_RDKIT_VERSION_OBJ < packaging.version.parse(rdkit_version)


def is_greater_than_current_rdkit_version(rdkit_version: str):
    return CURRENT_RDKIT_VERSION_OBJ > packaging.version.parse(rdkit_version)


def is_lower_eq_than_current_rdkit_version(rdkit_version: str):
    return CURRENT_RDKIT_VERSION_OBJ <= packaging.version.parse(rdkit_version)


def is_greater_eq_than_current_rdkit_version(rdkit_version: str):
    return CURRENT_RDKIT_VERSION_OBJ >= packaging.version.parse(rdkit_version)



import pytest

import pathlib

import fsspec
import datamol as dm


def test_copy_files(tmp_path):
    source_path = tmp_path / "source.txt"
    destination_path = tmp_path / "destination.txt"

    content = "hello this is a content"
    with open(source_path, "w") as f:
        f.write(content)

    dm.utils.fs.copy_file(source_path, destination_path)

    with open(destination_path) as f:
        f.read() == content


def test_copy_dir(tmp_path):
    source_path = tmp_path / "source_dir"
    source_path_subdir = source_path / "a_subdir"
    destination_path = tmp_path / "destination_dir"
    destination_path_subdir = destination_path / "a_subdir"

    dm.utils.fs.mkdir(source_path)
    dm.utils.fs.mkdir(source_path_subdir)

    content = "hello this is a content"
    file1_path = source_path / "hello.txt"
    with open(file1_path, "w") as f:
        f.write(content)

    file2_path = source_path_subdir / "hello.txt"
    with open(file2_path, "w") as f:
        f.write(content)

    assert not dm.utils.fs.is_dir(destination_path_subdir)
    assert not dm.utils.fs.is_dir(destination_path)

    dm.utils.fs.copy_dir(source_path, destination_path)

    assert dm.utils.fs.is_dir(destination_path_subdir)
    assert dm.utils.fs.is_dir(destination_path)
    assert dm.utils.fs.is_file(file1_path)
    assert dm.utils.fs.is_file(file2_path)

    with open(file1_path) as f:
        f.read() == content

    with open(file2_path) as f:
        f.read() == content


def test_mkdir(tmp_path):
    source_path = tmp_path / "source_dir"
    source_path_subdir = source_path / "a_subdir"

    dm.utils.fs.mkdir(source_path)

    assert dm.utils.fs.is_dir(source_path)
    assert not dm.utils.fs.is_dir(source_path_subdir)

    dm.utils.fs.mkdir(source_path_subdir)

    assert dm.utils.fs.is_dir(source_path)
    assert dm.utils.fs.is_dir(source_path_subdir)


@pytest.mark.skip_platform("win")
def test_cache_dir():
    cache_dir = dm.utils.fs.get_cache_dir("my_app")
    assert str(cache_dir).endswith("my_app")
    assert cache_dir.exists()
    assert cache_dir.is_dir()

    cache_dir = dm.utils.fs.get_cache_dir("my_app", suffix="likelydonotalreadyexist", create=False)
    assert str(cache_dir).endswith("likelydonotalreadyexist")
    assert not cache_dir.exists()
    assert not cache_dir.is_dir()

    cache_dir = dm.utils.fs.get_cache_dir("my_app", suffix="iamasuffix")
    assert str(cache_dir).endswith("iamasuffix")
    assert "my_app" in str(cache_dir)
    assert cache_dir.exists()
    assert cache_dir.is_dir()


def test_get_mapper(tmp_path):
    fsmapper = dm.utils.fs.get_mapper(str(tmp_path / "test.txt"))
    assert fsmapper.fs.protocol == "file"


@pytest.mark.skip_platform("win")
def test_get_basename(tmp_path):
    assert dm.utils.fs.get_basename(str(tmp_path / "test.txt")) == "test.txt"
    assert dm.utils.fs.get_basename("s3://a-bucket-that-likely-do-not-exist/test.txt") == "test.txt"


def test_get_extension(tmp_path):
    assert dm.utils.fs.get_extension(str(tmp_path / "test.txt")) == "txt"
    assert dm.utils.fs.get_extension("s3://a-bucket-that-likely-do-not-exist/test.txt") == "txt"


def test_exists(tmp_path):
    tmp_file = tmp_path / "test.txt"

    assert not dm.utils.fs.exists(tmp_file)
    assert not dm.utils.fs.is_file(tmp_file)

    assert dm.utils.fs.is_dir(tmp_path)
    assert not dm.utils.fs.is_dir(tmp_path / "likely-does-not-exist")

    with open(tmp_file, "w") as f:
        f.write("hello")

    assert dm.utils.fs.exists(tmp_file)
    assert dm.utils.fs.is_file(tmp_file)

    assert not dm.utils.fs.is_file(open(tmp_file))
    assert not dm.utils.fs.is_dir(open(tmp_file))


def test_get_protocol(tmp_path):
    assert dm.utils.fs.get_protocol(tmp_path / "ahahah.txt") == "file"
    assert dm.utils.fs.get_protocol("s3://a-bucket-that-likely-do-not-exist/test.txt") == "s3"


def test_is_local_path(tmp_path):
    assert dm.utils.fs.is_local_path(tmp_path / "ahahah.txt")
    assert not dm.utils.fs.is_local_path("s3://a-bucket-that-likely-do-not-exist/test.txt")


@pytest.mark.skip_platform("win")
def test_join(tmp_path):
    assert (
        dm.utils.fs.join("s3://a-bucket-that-likely-do-not-exist", "test.txt")
        == "s3://a-bucket-that-likely-do-not-exist/test.txt"
    )
    assert dm.utils.fs.join(tmp_path, "test.txt") == str(tmp_path / "test.txt")


def test_get_size(tmp_path):
    tmp_file = tmp_path / "test.txt"

    with open(tmp_file, "w") as f:
        f.write("hello")

    assert dm.utils.fs.get_size(tmp_file) > 0
    assert dm.utils.fs.get_size(open(tmp_file)) > 0
    assert dm.utils.fs.get_size(fsspec.open(tmp_file)) > 0


def test_md5(tmp_path):
    tmp_file = tmp_path / "test.txt"

    with open(tmp_file, "w") as f:
        f.write("hello")

    assert dm.utils.fs.md5(tmp_file) == "5d41402abc4b2a76b9719d911017c592"


@pytest.mark.skip_platform("win")
def test_glob(tmp_path):
    for i in range(5):
        tmp_file = tmp_path / f"test_{i}.txt"

        with open(tmp_file, "w") as f:
            f.write("hello")

    assert len(dm.utils.fs.glob(tmp_path / "*.txt")) == 5


def test_copy_file(tmp_path):
    tmp_file = tmp_path / "test.txt"

    assert dm.utils.fs.is_dir(tmp_path)
    assert dm.utils.fs.is_dir(str(tmp_path))
    assert dm.utils.fs.is_dir(pathlib.Path(str(tmp_path)))

    assert not dm.utils.fs.is_dir(tmp_path / "not_exist_dir")
    assert not dm.utils.fs.is_dir(str(tmp_path / "not_exist_dir"))
    assert not dm.utils.fs.is_dir(pathlib.Path(str(tmp_path / "not_exist_dir")))

    with open(tmp_file, "w") as f:
        f.write("hello")

    tmp_file2 = tmp_path / "test2.txt"
    assert not dm.utils.fs.is_file(tmp_file2)
    assert not dm.utils.fs.is_file(str(tmp_file2))
    assert not dm.utils.fs.is_file(pathlib.Path(str(tmp_file2)))

    dm.utils.fs.copy_file(tmp_file, tmp_file2)

    assert dm.utils.fs.is_file(tmp_file2)
    assert dm.utils.fs.is_file(str(tmp_file2))
    assert dm.utils.fs.is_file(pathlib.Path(str(tmp_file2)))
    assert open(tmp_file2).read() == "hello"

    with pytest.raises(ValueError):
        dm.utils.fs.copy_file(tmp_file, tmp_file2)

    tmp_file3 = tmp_path / "test3.txt"
    dm.utils.fs.copy_file(tmp_file, tmp_file3, progress=True)
    assert dm.utils.fs.is_file(tmp_file3)
    assert dm.utils.fs.is_file(str(tmp_file3))
    assert dm.utils.fs.is_file(pathlib.Path(str(tmp_file3)))
    assert open(tmp_file3).read() == "hello"



import datamol as dm


def test_find_mcs():
    smiles_list = [
        "C=CC(=O)NCCOc1cc2ncnc(Nc3ccc(Br)cc3F)c2cc1NC(=O)C=C",
        "C=CC(=O)Nc1cc2c(Nc3ccc(F)c(Br)c3)ncnc2cc1OCCCN1CCOCC1",
        "C=CC(=O)Nc1cc2c(Nc3ccc(Br)cc3F)ncnc2cc1OCCCNC(=O)CN(C)C",
        "C=CC(=O)Nc1cc2c(Nc3ccc(Br)cc3F)ncnc2cc1OCCNC(=O)NCC",
        "C=CC(=O)Nc1cc2c(Nc3ccc(Br)cc3F)ncnc2cc1OCCNC(=O)CN(C)C",
    ]
    mols = [dm.to_mol(s) for s in smiles_list]
    smarts = dm.find_mcs(mols=mols, timeout=2)

    # Load/export SMARTS to check RDKit versions compatibility.
    excepted_smarts = "[#6&!R]-&!@[#6&!R]-&!@[#8&!R]-&!@[#6&R]1:&@[#6&R]:&@[#6&R]2:&@[#7&R]:&@[#6&R]:&@[#7&R]:&@[#6&R](:&@[#6&R]:&@2:&@[#6&R]:&@[#6&R]:&@1-&!@[#7&!R]-&!@[#6&!R](=&!@[#8&!R])-&!@[#6&!R]=&!@[#6&!R])-&!@[#7&!R]-&!@[#6&R]1:&@[#6&R]:&@[#6&R]:&@[#6&R]:&@[#6&R]:&@[#6&R]:&@1"
    excepted_smarts_mol = dm.from_smarts(excepted_smarts)
    excepted_smarts = dm.to_smarts(excepted_smarts_mol)

    print(smarts)

    assert smarts == excepted_smarts



import pytest
import io

from rdkit import Chem

import datamol as dm
import numpy as np
import pandas as pd


def test_read_csv(datadir):
    data_path = datadir / "freesolv.csv"
    df = dm.read_csv(data_path)
    assert df.shape == (642, 4)
    assert list(df.columns) == ["iupac", "smiles", "expt", "calc"]


def test_read_csv_with_mol_col(datadir):
    data_path = datadir / "freesolv.csv"
    df = dm.read_csv(data_path, smiles_column="smiles", mol_column="mol")
    assert df.shape == (642, 5)
    assert set(df.columns) == {"iupac", "smiles", "expt", "calc", "mol"}
    assert isinstance(df.iloc[0]["mol"], Chem.rdchem.Mol)


def test_read_excel(datadir):
    data_path = datadir / "freesolv.xlsx"
    df = dm.read_excel(data_path, engine="openpyxl")
    assert df.shape == (642, 4)
    assert set(df.columns) == {"iupac", "smiles", "expt", "calc"}


def test_read_excel_with_mol_col(datadir):
    data_path = datadir / "freesolv.xlsx"
    df = dm.read_excel(data_path, smiles_column="smiles", mol_column="mol")
    assert df.shape == (642, 5)
    assert set(df.columns) == {"iupac", "smiles", "expt", "calc", "mol"}
    assert isinstance(df.iloc[0]["mol"], Chem.rdchem.Mol)


def test_read_sdf(datadir):
    data_path = datadir / "TUBB3-observations.sdf"

    # Read all sdf
    mols = dm.read_sdf(data_path)
    assert len(mols) == 10
    for mol in mols:
        assert isinstance(mol, Chem.rdchem.Mol)

    # Read 5 molecules
    mols = dm.read_sdf(data_path, max_num_mols=5)
    assert len(mols) == 5
    for mol in mols:
        assert isinstance(mol, Chem.rdchem.Mol)

    # Read more than the max number of mols in the file
    mols = dm.read_sdf(data_path, max_num_mols=111)
    assert len(mols) == 10
    for mol in mols:
        assert isinstance(mol, Chem.rdchem.Mol)

    data_path = datadir / "TUBB3-observations-last-broken.sdf"

    # Read all sdf with last mol being broken
    mols = dm.read_sdf(data_path)
    assert len(mols) == 9
    for mol in mols:
        assert isinstance(mol, Chem.rdchem.Mol)

    # Read all sdf with last mol being broken
    mols = dm.read_sdf(data_path, discard_invalid=False)
    assert len(mols) == 10
    for mol in mols[:-1]:
        assert isinstance(mol, Chem.rdchem.Mol)
    assert mols[-1] is None

    # Read all sdf with last mol being broken
    df = dm.read_sdf(data_path, discard_invalid=False, as_df=True, mol_column="mols")
    assert len(mols) == 10
    for mol in df["mols"].iloc[:-1]:
        assert isinstance(mol, Chem.rdchem.Mol)
    assert df["mols"].iloc[-1] is None


def test_read_sdf_gz(datadir):
    data_path = datadir / "TUBB3-observations.sdf.gz"
    mols = dm.read_sdf(data_path)
    assert len(mols) == 10
    for mol in mols:
        assert isinstance(mol, Chem.rdchem.Mol)


def test_read_sdf_as_df(datadir):
    data_path = datadir / "TUBB3-observations.sdf"
    df = dm.read_sdf(data_path, as_df=True)
    assert df.shape == (10, 12)
    assert set(df.columns) == {
        "smiles",
        "zinc_id",
        "ortholog_name",
        "gene_name",
        "affinity",
        "chembldocid",
        "title",
        "reference.pubmed_id",
        "reference.doi",
        "reference.chembl_id",
        "reference.journal",
        "reference.year",
    }


def test_read_sdf_as_df_parallel(datadir):
    data_path = datadir / "TUBB3-observations.sdf"
    df = dm.read_sdf(data_path, as_df=True, n_jobs=1)
    df2 = dm.read_sdf(data_path, as_df=True, n_jobs=-1)
    pd.testing.assert_frame_equal(df, df2)


def test_read_sdf_as_df_mol_col(datadir):
    data_path = datadir / "TUBB3-observations.sdf"
    df = dm.read_sdf(data_path, as_df=True, mol_column="mol")
    assert "mol" in df.columns
    assert isinstance(df.iloc[0]["mol"], Chem.rdchem.Mol)


def test_read_sdf_gz_as_df(datadir):
    data_path = datadir / "TUBB3-observations.sdf.gz"
    df = dm.read_sdf(data_path, as_df=True)

    assert df.shape == (10, 12)
    assert set(df.columns) == {
        "smiles",
        "zinc_id",
        "ortholog_name",
        "gene_name",
        "affinity",
        "chembldocid",
        "title",
        "reference.pubmed_id",
        "reference.doi",
        "reference.chembl_id",
        "reference.journal",
        "reference.year",
    }


def test_to_sdf(datadir, tmp_path):
    data_path = datadir / "TUBB3-observations.sdf.gz"

    df = dm.read_sdf(data_path, as_df=True)

    sdf_path = tmp_path / "mols.sdf"
    dm.to_sdf(df, sdf_path, smiles_column="smiles")

    new_df = dm.read_sdf(sdf_path, as_df=True)
    assert df.equals(new_df)


def test_to_sdf_mols(datadir, tmp_path):
    data_path = datadir / "TUBB3-observations.sdf.gz"

    mols = dm.read_sdf(data_path, as_df=False)

    sdf_path = tmp_path / "mols.sdf"
    dm.to_sdf(mols, sdf_path)

    new_mols = dm.read_sdf(sdf_path, as_df=False)
    assert [dm.to_smiles(mol) for mol in mols] == [dm.to_smiles(mol) for mol in new_mols]


def test_to_from_text(tmp_path):
    temp_file = tmp_path / "mols.smi"

    smiles_list = [
        "Cn1c(=S)ccc2nc[nH]c21",
        "Clc1n[nH]c2c1=[NH+]C(c1ccc[nH+]c1)C[NH+]=2",
        "Fc1ccsc1",
        "N#Cc1cc2c(o1)[NH2+]CCN2Cn1cnc2c1CSCC2",
        "O=CN1CCC2NC=CC2C1",
        "Oc1[nH]nc2c1-n1ncnc1C2",
        "OC1=NNC2(OC=CCO2)C2(C3CCCc4nonc43)NN=NN12",
        "[NH-]Sc1cc2nc[nH+]cc2o1",
        "[NH3+]C12CNCCOC1(N1CCCCC1)C=C(F)NC2",
    ]
    mols = [dm.to_mol(m) for m in smiles_list]

    # Save from text and read from text
    dm.to_smi(mols, temp_file)
    loaded_mols = dm.read_smi(temp_file)
    loaded_smiles = [dm.to_smiles(m) for m in loaded_mols]
    assert loaded_smiles == smiles_list

    # Check error raised when list is empty
    with pytest.raises(ValueError):
        dm.to_smi([], temp_file, error_if_empty=True)

    temp_file.unlink()

    # Check file like object works too
    file_like = io.StringIO()
    dm.to_smi(mols, file_like)
    assert file_like.getvalue().strip().split("\n") == smiles_list


def test_to_sdf_single_mol(tmp_path):
    sdf_path = tmp_path / "test.sdf"

    smiles = "CC1(C2C(C3C(C(=O)C(=C(C3(C(=O)C2=C(C4=C1C=CC=C4O)O)O)O)C(=O)N)N(C)C)O)O"
    mol = dm.to_mol(smiles)
    dm.to_sdf(mol, sdf_path)

    mols = dm.read_sdf(sdf_path)
    assert dm.to_smiles(mol) == dm.to_smiles(mols[0])


def test_sdf_props_and_conformer_preserved(tmp_path):
    sdf_path = tmp_path / "test.sdf"

    # Generate an SDF file
    props = dict(test_int=588, test_str="hello")
    smiles = "CC1(C2C(C3C(C(=O)C(=C(C3(C(=O)C2=C(C4=C1C=CC=C4O)O)O)O)C(=O)N)N(C)C)O)O"

    mol = dm.to_mol(smiles)
    mol = dm.set_mol_props(mol, props)
    mol = dm.conformers.generate(mol, n_confs=1)
    pos = mol.GetConformer().GetPositions()
    dm.to_sdf(mol, sdf_path)

    # Read sdf file
    mols = dm.read_sdf(sdf_path)
    mol = mols[0]

    # Check properties
    assert mol.GetPropsAsDict() == props

    # Check conformer
    conf = mol.GetConformer()
    assert mol.GetNumConformers() == 1
    assert conf.Is3D()
    np.testing.assert_almost_equal(conf.GetPositions(), pos, decimal=4)


def test_read_mol2(datadir):
    data_path = datadir / "test.mol2"

    # to list of mols
    mols = dm.read_mol2file(data_path)

    assert isinstance(mols[0], Chem.rdchem.Mol)
    assert isinstance(mols[1], Chem.rdchem.Mol)
    assert isinstance(mols[2], Chem.rdchem.Mol)
    # cases where mol2 formats are damaged
    assert mols[3] is None
    assert mols[4] is None
    assert mols[5] is None
    assert mols[6] is None
    assert mols[7] is None

    firstMol = dm.to_mol("c1ccncc1")
    secondMol = dm.to_mol("c1c[nH]cn1")

    assert dm.same_mol(mols[0], firstMol)
    assert dm.same_mol(mols[1], secondMol)
    assert dm.same_mol(mols[2], secondMol)

    # a case where exception is raised because of None values
    with pytest.raises(ValueError):
        mols = dm.read_mol2file(data_path, fail_if_invalid=True)


def test_read_save_molblock():
    mol = dm.to_mol("Cn1c(=O)c2c(ncn2C)n(C)c1=O")

    # to molblock
    molblock = dm.to_molblock(mol)

    assert isinstance(molblock, str)
    assert "END" in molblock
    assert "V2000" in molblock
    assert "RDKit" in molblock

    # read molblock
    mol2 = dm.read_molblock(molblock)
    assert dm.same_mol(mol, mol2)


def test_read_molblock_invalid():
    mol = dm.read_molblock("hello")
    assert mol is None

    with pytest.raises(ValueError):
        dm.read_molblock("hello", fail_if_invalid=True)


def test_to_xlsx(tmp_path):
    excel_path1 = tmp_path / "test1.xlsx"
    excel_path2 = tmp_path / "test2.xlsx"

    data = dm.freesolv()
    data = data.iloc[:10]
    data["mol"] = data["smiles"].apply(dm.to_mol)

    # write from df
    dm.to_xlsx(data, excel_path1)
    assert excel_path1.exists()

    # write from list of molecules
    mols = dm.from_df(data)
    dm.to_xlsx(mols, excel_path2)
    assert excel_path2.exists()


def test_to_xlsx_empty():
    mols = [None]
    with pytest.raises(ValueError):
        dm.to_xlsx(mols, "/dev/null")  # type: ignore


def test_read_pdbblock():
    pdbblock = """HETATM    1  N1  UNL     1       3.707  -1.649   1.733  1.00  0.00           N
HETATM    2  C1  UNL     1       2.987  -1.188   0.938  1.00  0.00           C
HETATM    3  C2  UNL     1       1.869  -0.581   0.250  1.00  0.00           C
HETATM    4  C3  UNL     1       1.632   0.783   0.355  1.00  0.00           C
HETATM    5  C4  UNL     1       0.344   1.270   0.178  1.00  0.00           C
HETATM    6  C5  UNL     1      -0.755   0.429   0.049  1.00  0.00           C
HETATM    7  N2  UNL     1      -1.944   0.727   0.745  1.00  0.00           N
HETATM    8  C6  UNL     1      -3.233   0.937   0.173  1.00  0.00           C
HETATM    9  C7  UNL     1      -4.177   1.582   1.134  1.00  0.00           C
HETATM   10  C8  UNL     1      -5.125   2.560   0.500  1.00  0.00           C
HETATM   11  C9  UNL     1      -4.297   3.560  -0.257  1.00  0.00           C
HETATM   12  C10 UNL     1      -3.440   2.952  -1.324  1.00  0.00           C
HETATM   13  C11 UNL     1      -3.253   1.476  -1.208  1.00  0.00           C
HETATM   14  N3  UNL     1      -0.470  -0.872  -0.207  1.00  0.00           N
HETATM   15  C12 UNL     1       0.804  -1.335  -0.226  1.00  0.00           C
HETATM   16  N4  UNL     1       0.815  -2.626  -0.634  1.00  0.00           N
HETATM   17  C13 UNL     1      -0.473  -3.035  -0.715  1.00  0.00           C
HETATM   18  C14 UNL     1      -0.971  -4.331  -0.663  1.00  0.00           C
HETATM   19  C15 UNL     1      -2.225  -4.591  -0.133  1.00  0.00           C
HETATM   20  C16 UNL     1      -3.085  -3.511   0.005  1.00  0.00           C
HETATM   21  C17 UNL     1      -2.511  -2.281   0.175  1.00  0.00           C
HETATM   22  C18 UNL     1      -1.271  -1.954  -0.359  1.00  0.00           C
HETATM   23  C19 UNL     1       2.789   1.645   0.539  1.00  0.00           C
HETATM   24  C20 UNL     1       3.510   2.001  -0.595  1.00  0.00           C
HETATM   25  C21 UNL     1       4.873   2.200  -0.578  1.00  0.00           C
HETATM   26  C22 UNL     1       5.560   2.162   0.605  1.00  0.00           C
HETATM   27  C23 UNL     1       4.838   1.985   1.767  1.00  0.00           C
HETATM   28  C24 UNL     1       3.502   1.683   1.730  1.00  0.00           C
CONECT    1    2    2    2
CONECT    2    3
CONECT    3    4    4   15
CONECT    4    5   23
CONECT    5    6    6
CONECT    6    7   14
CONECT    7    8
CONECT    8    9   13
CONECT    9   10
CONECT   10   11
CONECT   11   12
CONECT   12   13
CONECT   14   15   22
CONECT   15   16   16
CONECT   16   17
CONECT   17   18   18   22
CONECT   18   19
CONECT   19   20   20
CONECT   20   21
CONECT   21   22   22
CONECT   23   24   24   28
CONECT   24   25
CONECT   25   26   26
CONECT   26   27
CONECT   27   28   28
END"""

    mol = dm.read_pdbblock(pdbblock)

    print(dm.to_smiles(mol))

    assert mol is not None
    assert mol.GetNumAtoms() == 28
    assert dm.to_inchikey(mol) == "ZVAMKEUGOUZEJZ-UHFFFAOYSA-N"

    conf = mol.GetConformer()
    assert conf.Is3D()


def test_to_pdbblock():
    mol = dm.to_mol("N#Cc1c(cc(NC2CCCCC2)n2c1nc1ccccc21)-c1ccccc1")
    molblock = dm.to_pdbblock(mol)

    assert "HETATM" in molblock
    assert "CONECT" in molblock
    assert "C19" in molblock
    assert "C11" in molblock

    mol2 = dm.read_pdbblock(molblock)
    assert dm.to_inchikey(mol2) == dm.to_inchikey(mol)


def test_read_pdbfile(tmp_path):
    mol = dm.to_mol("N#Cc1c(cc(NC2CCCCC2)n2c1nc1ccccc21)-c1ccccc1")
    mol = dm.conformers.generate(mol, n_confs=1)

    pdb_path = tmp_path / "test.pdb"
    dm.to_pdbfile(mol, pdb_path)

    mol2 = dm.read_pdbfile(pdb_path)
    assert dm.to_inchikey(mol2) == dm.to_inchikey(mol)


def test_to_pdbfile(tmp_path):
    mol = dm.to_mol("N#Cc1c(cc(NC2CCCCC2)n2c1nc1ccccc21)-c1ccccc1")
    mol = dm.conformers.generate(mol, n_confs=1)

    pdb_path = tmp_path / "test.pdb"

    dm.to_pdbfile(mol, pdb_path)

    with open(pdb_path) as f:
        molblock = f.read()

    assert "HETATM" in molblock
    assert "CONECT" in molblock
    assert "C19" in molblock
    assert "C11" in molblock



from typing import Optional, Union
from typing import List
from typing import Optional

import numpy as np

from rdkit import Chem

from ..types import Mol
from ..utils.jobs import JobRunner
from ..utils import decorators
from ..mol import PERIODIC_TABLE
from ..mol import copy_mol


@decorators.disable_on_os("win")
def sasa(
    mol: Mol,
    conf_id: Optional[Union[int, List[int]]] = None,
    n_jobs: int = 1,
) -> np.ndarray:
    """Compute Solvent Accessible Surface Area of all the conformers
    using FreeSASA (https://freesasa.github.io/). Values are returned
    as an array and also stored within each conformer as a property
    called `rdkit_free_sasa`.

    Example:

    ```python
    smiles = "O=C(C)Oc1ccccc1C(=O)O"
    mol = dm.to_mol(smiles)
    mol = dm.conformers.generate(mol)

    # Compute SASA for all the conformers without parallelization
    sasa_values = dm.conformers.sasa(mol, conf_id=None, n_jobs=1)

    # If minimization has been enabled (default to True)
    # you can access the computed energy.
    conf = mol.GetConformer(0)
    props = conf.GetPropsAsDict()
    print(props)
    # {'rdkit_uff_energy': 1.7649408317784008}
    ```

    Args:
        mol: a molecule
        conf_id: Id of the conformers to compute. If None, compute all.
        n_jobs: Number of jobs for parallelization. Set to 1 to disable
            and -1 to use all cores.

    Returns:
        mol: the molecule with the conformers.
    """
    from rdkit.Chem import rdFreeSASA

    if mol.GetNumConformers() == 0:
        raise ValueError(
            "The molecule has 0 conformers. You can generate conformers with `dm.conformers.generate(mol)`."
        )

    # Get Van der Waals radii (angstrom)
    radii = [PERIODIC_TABLE.GetRvdw(atom.GetAtomicNum()) for atom in mol.GetAtoms()]

    # Which conformers to compute
    conf_ids = []
    if conf_id is None:
        # If None compute for all the conformers
        conf_ids = list(range(mol.GetNumConformers()))  # type: ignore
    elif isinstance(conf_id, int):
        conf_ids = [conf_id]
    else:
        conf_ids = conf_id

    # Compute solvent accessible surface area
    def _get_sasa(i):
        conf = mol.GetConformer(i)
        sasa = rdFreeSASA.CalcSASA(mol, radii, confIdx=conf.GetId())
        conf.SetDoubleProp("rdkit_free_sasa", sasa)
        return sasa

    runner = JobRunner(n_jobs=n_jobs)
    sasa_values = runner(_get_sasa, conf_ids)
    return np.array(sasa_values)


def get_coords(mol: Mol, conf_id: int = -1):
    """Get the coordinate of a conformer of a molecule.

    Args:
        mol: a molecule.
        conf_id: a conformer id.
    """

    if mol.GetNumConformers() == 0:
        raise ValueError("Molecule does not have any conformers.")

    conf = mol.GetConformer(id=conf_id)
    return conf.GetPositions()


def center_of_mass(
    mol: Mol,
    use_atoms: bool = True,
    digits: Optional[int] = None,
    conf_id: int = -1,
) -> np.ndarray:
    """Compute the center of mass of a conformer of a molecule.

    Args:
        mol: a molecule
        use_atoms: Whether to compute the true center of mass or the geometrical center.
        digits: Number of digits to round to.
        conf_id: the conformer id.

    Returns
        cm: Center of mass or geometrical center
    """
    coords = get_coords(mol, conf_id=conf_id)
    atom_weight = np.ones((coords.shape[0]))

    if use_atoms:
        atom_weight = np.array([atom.GetMass() for atom in mol.GetAtoms()])

    atom_weight = atom_weight[:, None]
    atom_weight /= atom_weight.sum()
    center = (coords * atom_weight).sum(axis=0)

    if digits is not None:
        center = center.round(digits)

    return center


def keep_conformers(
    mol: Mol,
    indices_to_keep: Union[int, List[int]] = -1,
    assign_id: bool = True,
    copy: bool = True,
):
    """Keep on the specified conformer(s) in `indices_to_keep`.

    Args:
        mol: A molecule.
        indices_to_keep: A indice or a least of indices of conformers to keep.
        assign_id: Whether to assign the kept conformers an id or keep the original one.
        copy: Whether to copy the molecule or not.
    """

    if copy:
        mol = copy_mol(mol)

    if not isinstance(indices_to_keep, list):
        indices_to_keep = [indices_to_keep]

    # Extract conformers to keep
    confs_to_keep = [mol.GetConformer(conf_id) for conf_id in indices_to_keep]

    # Copy current mol and remove all conformers
    mol2 = copy_mol(mol)
    mol2.RemoveAllConformers()

    # Add conformers
    _ = [mol2.AddConformer(conf, assignId=assign_id) for conf in confs_to_keep]

    # Cleanup
    mol = mol2

    return mol



import pytest

import datamol as dm
import numpy as np


def test_esol():
    smiles = "Nc1cnn(-c2ccccc2)c(=O)c1Cl"
    mol = dm.to_mol(smiles)

    assert np.allclose(dm.predictors.esol(mol), -2.627091966265316)


def test_esol_from_data():
    data = dm.freesolv()
    data = data.iloc[:20]

    with pytest.raises(KeyError):
        dm.predictors.esol_from_data(data)

    data["mol"] = data["smiles"].apply(dm.to_mol)
    data["clogp"] = data["mol"].apply(dm.descriptors.clogp)
    data["mw"] = data["mol"].apply(dm.descriptors.mw)
    data["n_rotatable_bonds"] = data["mol"].apply(dm.descriptors.n_rotatable_bonds)
    data["n_aromatic_atoms_proportion"] = data["mol"].apply(
        dm.descriptors.n_aromatic_atoms_proportion
    )

    # dataframe
    esol_values = dm.predictors.esol_from_data(data)
    assert esol_values.dtype == float
    assert esol_values.shape == (20,)

    # series
    v = dm.predictors.esol_from_data(data.iloc[0])
    v = float(v)
    assert type(v) == float

    # dict
    v = dm.predictors.esol_from_data(data.iloc[0].to_dict())
    v = float(v)
    assert type(v) == float



from typing import Dict
from typing import List
from typing import Union
from typing import Optional

from rdkit.Chem.rdmolops import GetAdjacencyMatrix
from rdkit.Chem.AllChem import RenumberAtoms  # type: ignore
from loguru import logger

import datamol as dm


def _get_networkx():
    try:
        import networkx as nx

        return nx
    except ImportError:
        raise ImportError("You must install networkx from https://networkx.org/.")


def to_graph(mol: dm.Mol):
    """Convert a molecule to a network x graph. A list of properties are added
    to every nodes and edges.

    Args:
        mol (dm.Mol): a molecule.

    Returns:
        mol_graph (networkx.Graph): a graph representing the molecule.
    """

    nx = _get_networkx()

    mol_graph = nx.Graph()
    for atom in mol.GetAtoms():
        mol_graph.add_node(
            atom.GetIdx(),
            atomic_num=atom.GetAtomicNum(),
            formal_charge=atom.GetFormalCharge(),
            chiral_tag=atom.GetChiralTag(),
            hybridization=atom.GetHybridization(),
            num_explicit_hs=atom.GetNumExplicitHs(),
            implicit_valence=atom.GetImplicitValence(),
            degree=atom.GetDegree(),
            symbol=atom.GetSymbol(),
            ring_atom=atom.IsInRing(),
            is_aromatic=atom.GetIsAromatic(),
        )
    for bond in mol.GetBonds():
        mol_graph.add_edge(
            bond.GetBeginAtomIdx(),
            bond.GetEndAtomIdx(),
            bond_type=bond.GetBondType(),
        )
    return mol_graph


def get_all_path_between(
    mol: dm.Mol,
    atom_idx_1: int,
    atom_idx_2: int,
    ignore_cycle_basis: bool = False,
) -> list:
    """Get all simple path between two atoms of a molecule

    Args:
        mol (dm.Mol): a molecule
        atom_idx_1 (int): Atom index 1.
        atom_idx_2 (int): Atom index 2.
        ignore_cycle_basis: Whether to ignore cycle basis.
            Defaults to False.

    Returns:
        list of path between two atoms.
    """

    nx = _get_networkx()

    adj = GetAdjacencyMatrix(mol)
    G = nx.Graph(adj)
    path = nx.all_simple_paths(G, source=atom_idx_1, target=atom_idx_2)

    if ignore_cycle_basis:
        rings = [set(x) for x in mol.GetRingInfo().AtomRings()]
        final_path = []
        for p in path:
            reject_path = False
            for r in rings:
                if r.issubset(set(p)):
                    reject_path = True
                    break
            if not reject_path:
                final_path.append(p)
        path = final_path

    return list(path)


def match_molecular_graphs(
    mol1: dm.Mol,
    mol2: dm.Mol,
    match_atoms_on: List[str] = ["atomic_num"],
    match_bonds_on: List[str] = ["bond_type"],
) -> List[Dict[int, int]]:
    """
    Match the node indices of 2 molecular graphs,
    with optional usage of atomic number and edge type.

    Note:
        The matching fails if the hydrogens are implicit in one molecule,
        but explicit in the other.

    Note:
        Explicit hydrogens might lead to too many matches, since for an atom with 2
        hydrogens, they can be re-ordered in any way.

    Args:
        mol1: A molecule.
        mol2: A molecule.
        match_atoms_on: Properties on which to match the atom types.
            By default, it matches on the `'atomic_num'` property.
            Empty list means that it does not consider atom features during matching.

            Other properties are defined by the `datamol.graph.to_graph` function and include:
            - atomic_num
            - formal_charge
            - chiral_tag
            - hybridization
            - num_explicit_hs
            - implicit_valence
            - degree
            - symbol
            - ring_atom
            - is_aromatic

        match_bonds_on: Properties on which to match the bond types.
            Empty list means that it does not consider bond features during matching.
            By default, it matches on the `'bond_type'` property.
            No other properties are defined by the `datamol.graph.to_graph` function.

    Returns:
        A list of all matches dictionaries. In case of a single match, the list has len==1.
        Each dictionary contains as key the indices of `mol1` and as value the corresponding
        indices of `mol2`.
    """

    nx = _get_networkx()

    if isinstance(match_atoms_on, str):
        match_atoms_on = [match_atoms_on]
    if isinstance(match_bonds_on, str):
        match_bonds_on = [match_bonds_on]

    def node_match_fn(node1, node2):
        """Function that matches the atomic number"""
        return all([node1[prop] == node2[prop] for prop in match_atoms_on])

    def edge_match_fn(edge1, edge2):
        """Function that matches the bond type"""
        return all([edge1[prop] == edge2[prop] for prop in match_bonds_on])

    # Convert to networkx graph
    g1 = to_graph(mol1)
    g2 = to_graph(mol2)

    # Use the `match` function to find the matching indices
    node_match = node_match_fn if len(match_atoms_on) > 0 else None
    edge_match = edge_match_fn if len(match_bonds_on) > 0 else None
    graph_matcher = nx.algorithms.isomorphism.vf2userfunc.GraphMatcher(
        g1, g2, node_match=node_match, edge_match=edge_match
    )
    matches = list(graph_matcher.match())

    return matches


def reorder_mol_from_template(
    mol: dm.Mol,
    mol_template: dm.Mol,
    enforce_atomic_num: bool = False,
    enforce_bond_type: bool = False,
    ambiguous_match_mode: str = "No",
    verbose: bool = True,
) -> Optional[dm.Mol]:
    """
    Re-order the nodes of a molecular graph from the nodes of a template molecule.
    Molecular graphs and atom types need to be identical, but edge types and charges
    are not enforced.

    This is particularily useful when dealing with XYZ files containing node ordering,
    but with missing information regarding charges and edge types.

    !!! note

        * If you only need to match bond orders, you can check the function
        `rdkit.Chem.AllChem.AssignBondOrdersFromTemplate`.
        * The matching fails if the hydrogens are implicit in one molecule,
        but explicit in the other.
        * Explicit hydrogens might lead to too many matches, since for an atom with 2
        hydrogens, they can be re-ordered in any way.

    Args:
        mol: The molecule to re-order
        mol_template: The molecule containing the right node order.
        enforce_atomic_num: Whether to enforce atomic number. Atomic numbers are always enforced
            for a first try. If no match are found and this parameter is `False`,
            the matching is tried again.
        enforce_bond_type: Whether to enforce bond types. Bond types are always enforced
            for a first try. If no match are found and this parameter is `False`,
            the matching is tried again.
        ambiguous_match_mode: Whether to allow ambiguous matching. This means that,
            if there are many matches to the molecule, it will still re-order
            the molecule according to specific rules. Options are:
            - "no": Does not allow ambiguous matching.
            - "hs-only": Allow matching of ambiguous hydrogens. Does not work if trying
              to match implicit with explicit hydrogens.
            - "first": Return the first match.
            - "best": Return the match with the least errors on atom type, edges type, and edge stereo.
              Errors on the atoms are counted with 1 point, on the charge with 0.25 points,
              on the edges with 0.25 points, and on the Stereo with 0.05 points.
              If the option `enforce_atomic_num` is used, then no errors on the atoms are allowed.
              If the option `enforce_bond_type` is used, then no errors on the edges are allowed.
            - "best-first": "best", followed by "first".
        verbose: Whether to warn when the matching does not work or is ambiguous.
            Different warnings are raised depending on the value of `ambiguous_match_mode`.

    Returns:
        - `None` if the molecular graphs do not match (both the graph and atom types).
            Pring a warning.
        - `None` if multiple matche are found, which can happen for symmetric molecules such as benzene
            Pring a warning.
        - `Mol` The re-ordered molecule when a single match is found.
    """

    ambiguous_match_mode = ambiguous_match_mode.lower()

    # Match the ordering of the graphs
    matches = match_molecular_graphs(
        mol_template,
        mol,
        match_atoms_on=["atomic_num"],
        match_bonds_on=["bond_type"],
    )

    # If no matches were found, retry without bond types
    if (len(matches) == 0) and (not enforce_bond_type):
        matches = match_molecular_graphs(
            mol_template,
            mol,
            match_atoms_on=["atomic_num"],
            match_bonds_on=[],
        )

    # If no matches were found, retry without atom types
    if (len(matches) == 0) and (not enforce_atomic_num):
        matches = match_molecular_graphs(
            mol_template,
            mol,
            match_atoms_on=[],
            match_bonds_on=["bond_type"],
        )

    # If no matches were found, retry without bond and atom types
    if (len(matches) == 0) and (not enforce_bond_type) and (not enforce_atomic_num):
        matches = match_molecular_graphs(mol_template, mol, match_atoms_on=[], match_bonds_on=[])

    # If no match were found, exit the function and return None
    if len(matches) == 0:
        if verbose:
            logger.warning("No match was found")
        return None

    if len(matches) > 1:
        # In case we want to allow ambiguous match of hydrogens
        if ambiguous_match_mode == "hs-only":
            first_keys = list(matches[0].keys())
            all_hs_mismatch = True
            for this_match in matches:
                this_keys = list(this_match.keys())
                keys_mismatch = [
                    ii for ii in range(len(this_keys)) if (first_keys[ii] != this_keys[ii])
                ]
                atoms_mismatch = [mol.GetAtomWithIdx(key).GetAtomicNum() for key in keys_mismatch]
                all_hs = all([atom == 1 for atom in atoms_mismatch])
                if not all_hs:
                    all_hs_mismatch = False
                    break
            if all_hs_mismatch:
                matches = matches[0:1]
            else:
                if verbose:
                    logger.warning(
                        f"{len(matches)} matches were found, ordering is ambiguous, even when ignoring hydrogens"
                    )
                return None

        # Compute the number of atoms and bonds mismatch, and select the one with the least mismatch
        if (ambiguous_match_mode in ["best", "best-first"]) and not (
            enforce_atomic_num and enforce_bond_type
        ):
            num_mismatches = []
            for this_match in matches:
                num_atoms_mismatch, num_charge_mismatch = 0, 0

                # Get the number of atomic mismatch
                for key, val in this_match.items():
                    atom1 = mol.GetAtomWithIdx(val)
                    atom2 = mol_template.GetAtomWithIdx(key)
                    num_atoms_mismatch += atom1.GetAtomicNum() != atom2.GetAtomicNum()
                    num_charge_mismatch += atom1.GetFormalCharge() != atom2.GetFormalCharge()

                # Get the number of bond mismatch
                num_bonds_type_mismatch, num_bonds_stereo_mismatch = 0, 0
                for bond1 in mol_template.GetBonds():
                    begin_idx, end_idx = bond1.GetBeginAtomIdx(), bond1.GetEndAtomIdx()
                    bond2 = mol.GetBondBetweenAtoms(this_match[begin_idx], this_match[end_idx])
                    num_bonds_type_mismatch += bond1.GetBondType() != bond2.GetBondType()
                    num_bonds_stereo_mismatch += (bond1.GetStereo() != bond2.GetStereo()) or (
                        bond1.GetBondDir() != bond2.GetBondDir()
                    )

                num_mismatches.append(
                    (1 * num_atoms_mismatch)
                    + (0.25 * num_charge_mismatch)
                    + (0.25 * num_bonds_type_mismatch)
                    + (0.05 * num_bonds_stereo_mismatch)
                )
            min_mismatch_idx = [
                ii for ii in range(len(num_mismatches)) if num_mismatches[ii] == min(num_mismatches)
            ]
            matches = [matches[idx] for idx in min_mismatch_idx]

        # Select the first matching element
        if ambiguous_match_mode in ["first", "best-first"]:
            matches = [matches[0]]

    if len(matches) > 1:
        # If many matches were found, exit the function and return None
        if ambiguous_match_mode == "no":
            if verbose:
                logger.warning(f"{len(matches)} matches were found, ordering is ambiguous")
            return None

    # Re-order the molecule from the matching indices of the template
    match = matches[0]
    match = [match[ii] for ii in range(mol.GetNumAtoms())]
    reordered_mol = RenumberAtoms(mol, match)

    return reordered_mol



from typing import List
from typing import Callable
from typing import Optional
from typing import Sequence
from typing import Union
from typing import Tuple

import operator
import functools
from collections import defaultdict as ddict

from loguru import logger

import pandas as pd
import numpy as np
from scipy.spatial import distance

from rdkit.Chem import DataStructs
from rdkit.ML.Cluster import Butina

from rdkit.SimDivFilters.rdSimDivPickers import ClusterMethod
from rdkit.SimDivFilters.rdSimDivPickers import HierarchicalClusterPicker
from rdkit.SimDivFilters.rdSimDivPickers import MaxMinPicker
from rdkit.SimDivFilters.rdSimDivPickers import LeaderPicker

import datamol as dm

from .types import Mol


def cluster_mols(
    mols: Union[Sequence[Mol], pd.Series],
    cutoff: float = 0.2,
    feature_fn: Optional[Callable] = None,
    n_jobs: Optional[int] = 1,
):
    """Cluster a set of molecules using the butina clustering algorithm and a given threshold.

    Args:
        mols: a list of molecules.
        cutoff: Cuttoff for the clustering. Default to 0.2.
        feature_fn: A feature function that takes a Mol object
            and return molecular features. By default, the `dm.to_fp()` is used.
            Default to None.
        n_jobs: Number of jobs for parallelization. Let to 1 for no
            parallelization. Set to -1 to use all available cores.
    """

    if feature_fn is None:
        feature_fn = functools.partial(dm.to_fp, as_array=False)

    features = dm.parallelized(feature_fn, mols, n_jobs=n_jobs)

    dists = []
    n_mols = len(mols)

    for i in range(1, n_mols):
        dist = DataStructs.cDataStructs.BulkTanimotoSimilarity(
            features[i], features[:i], returnDistance=True
        )
        dists.extend([x for x in dist])

    # now cluster the data
    cluster_indices = Butina.ClusterData(dists, n_mols, cutoff, isDistData=True)
    cluster_mols = [operator.itemgetter(*cluster)(mols) for cluster in cluster_indices]

    # Make single mol cluster a list
    cluster_mols = [[c] if isinstance(c, Mol) else c for c in cluster_mols]

    return cluster_indices, cluster_mols


def pick_diverse(
    mols: List[Mol],
    npick: int,
    initial_picks: Optional[List[int]] = None,
    feature_fn: Optional[Callable] = None,
    dist_fn: Optional[Callable] = None,
    seed: int = 42,
    n_jobs: Optional[int] = 1,
) -> Tuple[int, list]:
    r"""Pick a set of diverse molecules based on they fingerprint.

    Args:
        mols: a list of molecules.
        npick: Number of element to pick from mols, including the preselection.
        initial_picks: Starting list of index for molecules that should be in the
            set of picked molecules. Default to None.
        feature_fn: A feature function that takes a Mol object
            and return molecular features. By default, the `dm.to_fp()` is used.
            Default to None.
        dist_fn: A function that takes two indexes (i,j) and return the
            distance between them. You might use partial to set the fingerprints as input.
            By default, the Tanimoto similarity will be used. Default to None.
        seed: seed for reproducibility
        n_jobs: Number of jobs for parallelization. Let to 1 for no
            parallelization. Set to -1 to use all available cores.

    Returns:
        picked_inds: index of the molecule that have been picked
        mols: molecules that have been picked
    """

    if feature_fn is None:
        feature_fn = functools.partial(dm.to_fp, as_array=False)

    features = dm.parallelized(feature_fn, mols, n_jobs=n_jobs)

    def distij(i, j, features=features):
        return 1.0 - DataStructs.cDataStructs.TanimotoSimilarity(features[i], features[j])

    if dist_fn is None:
        dist_fn = distij

    picker = MaxMinPicker()
    initial_picks = [] if initial_picks is None else initial_picks
    picked_inds = picker.LazyPick(dist_fn, len(mols), npick, firstPicks=initial_picks, seed=seed)
    picked_inds = np.array(picked_inds)
    picked_mols = [mols[x] for x in picked_inds]

    return picked_inds, picked_mols


def pick_centroids(
    mols: List[Mol],
    npick: int = 0,
    initial_picks: Optional[List[int]] = None,
    threshold: float = 0.5,
    feature_fn: Optional[Callable] = None,
    dist_fn: Optional[Callable] = None,
    seed: int = 42,
    method: str = "sphere",
    n_jobs: Optional[int] = 1,
) -> Tuple[int, list]:
    r"""Pick a set of `npick` centroids from a list of molecules.

    Args:
        mols: a list of molecules.
        npick: Number of element to pick from mols, including the preselection.
        threshold: Minimum distance between centroids for `maxmin` and sphere exclusion (`sphere`) methods.
        initial_picks: Starting list of index for molecules that should be in the
            set of picked molecules. Default to None.
        feature_fn (callable, optional): A feature function that takes a Mol object
            and return molecular features. By default, the `dm.to_fp()` is used.
            Default to None.
        dist_fn: A function that takes two indexes (i,j) and return the
            distance between them. You might use partial to set the fingerprints as input.
            By default, the Tanimoto similarity will be used. Default to None.
        seed: seed for reproducibility
        method: Picking method to use. One of  `sphere`, `maxmin` or any
            supported rdkit hierarchical clustering method such as `centroid`, `clink`, `upgma`
        n_jobs: Number of jobs for parallelization. Let to 1 for no
            parallelization. Set to -1 to use all available cores.

    Returns:
        picked_inds: index of the molecule that have been selected as centroids
        mols: molecules that have been picked
    """

    n_mols = len(mols)
    if feature_fn is None:
        feature_fn = functools.partial(dm.to_fp, as_array=False)

    features = dm.parallelized(feature_fn, mols, n_jobs=n_jobs)

    def distij(i, j, features=features):
        return 1.0 - DataStructs.cDataStructs.TanimotoSimilarity(features[i], features[j])

    if dist_fn is None:
        dist_fn = distij

    initial_picks = [] if initial_picks is None else initial_picks

    if method == "maxmin":
        picker = MaxMinPicker()
        picked_inds, _ = picker.LazyPickWithThreshold(
            dist_fn,
            n_mols,
            pickSize=npick,
            threshold=threshold,
            firstPicks=initial_picks,
            seed=seed,
        )

    elif method == "sphere":
        picker = LeaderPicker()
        picked_inds = picker.LazyPick(
            dist_fn, n_mols, threshold=threshold, pickSize=npick, firstPicks=initial_picks
        )

    elif method.upper() in ClusterMethod.names.keys() and npick:
        if initial_picks:
            logger.warning(
                "Initial picks is not supported by hierarchical clustering. You pick has been discarded."
            )

        dist_mat = dm.parallelized(
            distij, list(zip(*np.tril_indices(len(mols), k=-1))), arg_type="args"
        )
        dist_mat = np.asarray(dist_mat)
        picker = HierarchicalClusterPicker(ClusterMethod.names[method.upper()])
        picked_inds = picker.Pick(dist_mat, n_mols, npick)
    else:
        raise ValueError(f"Picking method {method} with {npick} elements to pick is not supported.")
    picked_inds = np.array(picked_inds)
    picked_mols = [mols[x] for x in picked_inds]

    return picked_inds, picked_mols


def assign_to_centroids(
    mols: List[Mol],
    centroids: List[Mol],
    feature_fn: Optional[Callable] = None,
    dist_fn: Optional[Callable] = None,
    n_jobs: Optional[int] = 1,
) -> Tuple[dict, list]:
    r"""Assign molecules to centroids. Each molecule will be assigned to the closest centroid.

    Args:
        mols: a list of molecules to assign to centroids
        centroids: list of molecules to use as centroid
        feature_fn: A feature function that takes a Mol object
            and return molecular features. By default, the `dm.to_fp()` is used.
            Default to None.
        dist_fn: A function that takes two indexes (i,j) and return the
            distance between them. You might use partial to set the fingerprints as input.
            By default, the Tanimoto similarity will be used. Default to None.
        n_jobs: Number of jobs for parallelization. Let to 1 for no
            parallelization. Set to -1 to use all available cores.

    Returns:
        clusters_map: dict of index mapping each centroid index to the molecule index in the cluster
        clusters_list: list of all molecules in each cluster. The cluster index follows the index of the centroid.
            Note that the centroid molecule is not added to the cluster.
    """

    if feature_fn is None:
        feature_fn = functools.partial(dm.to_fp, as_array=False)

    all_mols = [x for x in mols] + [c for c in centroids]
    features = dm.parallelized(feature_fn, all_mols, n_jobs=n_jobs)

    def distij(i, j, features=features):
        return 1.0 - DataStructs.cDataStructs.TanimotoSimilarity(features[int(i)], features[int(j)])

    if dist_fn is None:
        dist_fn = distij

    clusters_map = ddict(list)
    clusters_list = [[] for _ in centroids]
    query_inds = np.expand_dims(np.arange(len(mols), dtype=int), axis=1)
    centroid_inds = np.expand_dims(np.arange(len(centroids), dtype=int), axis=1) + len(mols)
    dist_mat = distance.cdist(query_inds, centroid_inds, metric=distij)
    closest = np.argmin(dist_mat, axis=1)
    for ind, cluster_ind in enumerate(closest):  # type: ignore
        clusters_map[cluster_ind].append(ind)
        clusters_list[cluster_ind].append(mols[ind])
    return clusters_map, clusters_list



import pytest

import pandas as pd
import datamol as dm


def test_template_align():
    data: pd.DataFrame = dm.cdk2(as_df=True)  # type: ignore
    data = data.iloc[:6].copy()  # type: ignore

    template = data.iloc[0]["mol"]
    data["aligned_mol"] = data["mol"].apply(lambda x: dm.align.template_align(x, template=template))
    assert data["aligned_mol"].apply(lambda x: isinstance(x, dm.Mol)).all() == True

    template = data.iloc[0]["smiles"]
    data["aligned_mol"] = data["smiles"].apply(
        lambda x: dm.align.template_align(x, template=template)
    )
    assert data["aligned_mol"].apply(lambda x: isinstance(x, dm.Mol)).all() == True

    template = data.iloc[0]["mol"]
    data["aligned_mol"] = data["mol"].apply(
        lambda x: dm.align.template_align(x, template=template, auto_select_coord_gen=True)
    )
    assert data["aligned_mol"].apply(lambda x: isinstance(x, dm.Mol)).all() == True

    template = data.iloc[0]["mol"]
    data["aligned_mol"] = data["mol"].apply(
        lambda x: dm.align.template_align(x, template=template, use_depiction=False)
    )
    assert data["aligned_mol"].apply(lambda x: isinstance(x, dm.Mol)).all() == True

    template = None
    data["aligned_mol"] = data["mol"].apply(lambda x: dm.align.template_align(x, template=template))
    assert data["aligned_mol"].apply(lambda x: isinstance(x, dm.Mol)).all() == True

    template = None
    data["aligned_mol"] = data["mol"].apply(
        lambda x: dm.align.template_align(x, template=template, copy=False)
    )
    assert data["aligned_mol"].apply(lambda x: isinstance(x, dm.Mol)).all() == True

    assert dm.align.template_align(None) is None


def test_auto_align_many():
    data: pd.DataFrame = dm.solubility(as_df=True)  # type: ignore
    data = data.iloc[:32].copy()  # type: ignore

    excepted_cluster_size = [17, 9, 5, 9, 9]

    for i, partition_method in enumerate(
        [
            "cluster",
            "scaffold",
            "anongraph-scaffold",
            "anon-scaffold",
            "strip-scaffold",
        ]
    ):
        print(partition_method)

        data["aligned_mol"] = dm.align.auto_align_many(
            data["mol"],
            partition_method=partition_method,
        )

        props = data["aligned_mol"].apply(lambda x: pd.Series(x.GetPropsAsDict()))

        assert "dm.auto_align_many.cluster_id" in props.columns
        assert "dm.auto_align_many.core" in props.columns
        assert props["dm.auto_align_many.cluster_id"].dtype.name == "int64"
        assert props["dm.auto_align_many.core"].dtype.name == "object"

        assert props["dm.auto_align_many.cluster_id"].unique().shape[0] == excepted_cluster_size[i]

    with pytest.raises(ValueError):
        dm.align.auto_align_many(data["mol"], partition_method="invalid")



import datamol as dm


def test_match_substructure():
    mol1 = dm.to_mol("CC(=O)OC1=CC=CC=C1C(=O)O")
    mol2 = dm.to_mol("CCN(CC)CC(=O)CC(C)NC1=C2C=CC(=CC2=NC=C1)Cl")

    query1 = dm.from_smarts("[C;H0](=O)")
    query2 = dm.to_mol("CN(C)")

    # Test multiple scenarios

    dm.viz.match_substructure(
        mols=[mol1, mol2],
        queries=[query1, query2],
        highlight_bonds=True,
        use_svg=True,
    )
    dm.viz.match_substructure(
        mols=mol1,
        queries=[query1, query2],
        highlight_bonds=True,
        use_svg=True,
    )
    dm.viz.match_substructure(
        mols=[mol1, mol2],
        queries=query1,
        highlight_bonds=False,
        use_svg=False,
    )
    dm.viz.match_substructure(
        mols=mol1,
        queries=query2,
        highlight_bonds=True,
        use_svg=False,
    )



import math
import numbers
import operator
import unittest

import numpy as np
import pandas as pd

from functools import reduce

import datamol as dm


def random_fn(*args, op="mul", **kwargs):
    """Perform random functions on a list"""
    all_values = [x for x in args if isinstance(x, numbers.Number)]
    all_values += [x for x in kwargs.values() if isinstance(x, numbers.Number)]
    op_fn = getattr(operator, op, None)
    if op_fn is None:
        op_fn = getattr(math, op)
        return op_fn(all_values[0])
    return reduce(op_fn, all_values)


class TestJobs(unittest.TestCase):
    def test_sequential(self):
        jobrunner = dm.JobRunner(n_jobs=None, progress=False)
        # practically do nothing (add a single value with nothing)
        o1 = jobrunner(random_fn, [9, 25, 1024], op="add")
        self.assertEqual(o1, [9, 25, 1024])

        # take the sqrt
        o2 = jobrunner(random_fn, [9, 25, 1024], op="sqrt")
        self.assertEqual(o2, [3, 5, 32])

        # multiply all inputs
        o3 = jobrunner(random_fn, [(1, 2, 3), (4, 5, 6), (3, 4, 0)], arg_type="args", op="mul")
        self.assertEqual(o3, [6, 4 * 5 * 6, 0])

        # do the same thing but with kwargs
        o4 = jobrunner(
            random_fn,
            iter([dict(a=1, b=2, c=3), dict(a=4, b=5, c=6), dict(a=3, b=4, c=0)]),
            arg_type="kwargs",
            op="mul",
        )
        self.assertEqual(o4, [6, 4 * 5 * 6, 0])

        o5 = jobrunner(random_fn, np.asarray([9, 25, 1024]), op="add")
        self.assertEqual(o5, [9, 25, 1024])

    def test_parallel(self):
        jobrunner1 = dm.JobRunner(n_jobs=4, progress=True)  # use loky backend
        o1 = jobrunner1(random_fn, [9, 25, 1024], op="add")
        self.assertEqual(o1, [9, 25, 1024])

        o5 = jobrunner1(random_fn, np.asarray([9, 25, 1024]), op="add")
        self.assertEqual(o5, [9, 25, 1024])

        o3 = jobrunner1(random_fn, [(1, 2, 3), (4, 5, 6), (3, 4, 0)], arg_type="args", op="mul")
        self.assertEqual(o3, [6, 4 * 5 * 6, 0])

        # use threads instead, no progress
        jobrunner2 = dm.JobRunner(n_jobs=2, progress=False, prefer="threads")
        o2 = jobrunner2(random_fn, [9, 25, 1024], op="sqrt")
        self.assertEqual(o2, [3, 5, 32])

        o4 = jobrunner2(
            random_fn,
            iter([dict(a=1, b=2, c=3), dict(a=4, b=5, c=6), dict(a=3, b=4, c=0)]),
            arg_type="kwargs",
            op="mul",
        )
        self.assertEqual(o4, [6, 4 * 5 * 6, 0])

    def test_seq_vs_parallel(self):
        # test parallel vs sequential
        jobrunner = dm.JobRunner(n_jobs=4, progress=False)  # use loky backend
        o_seq = jobrunner.sequential(
            random_fn, [(1, 2, 3), (4, 5, 6), (3, 4, 0)], arg_type="args", op="mul"
        )
        o_par = jobrunner.parallel(
            random_fn, [(1, 2, 3), (4, 5, 6), (3, 4, 0)], arg_type="args", op="mul"
        )
        self.assertEqual(o_seq, o_par)

    def test_parallelized(self):
        def fn(x):
            return x**2

        results = dm.parallelized(
            fn,
            [{"x": i} for i in range(10)],
            scheduler="processes",
            n_jobs=None,
            arg_type="kwargs",
            progress=True,
        )
        assert results == [0, 1, 4, 9, 16, 25, 36, 49, 64, 81]

        results = dm.parallelized(
            fn,
            [[i] for i in range(10)],
            scheduler="processes",
            n_jobs=None,
            arg_type="args",
            progress=True,
        )
        assert results == [0, 1, 4, 9, 16, 25, 36, 49, 64, 81]

        results = dm.parallelized(
            fn,
            range(10),
            scheduler="processes",
            n_jobs=None,
            progress=False,
        )
        assert results == [0, 1, 4, 9, 16, 25, 36, 49, 64, 81]

    def test_job_kwargs(self):
        def fn(x):
            return x**2

        results = dm.parallelized(
            fn,
            [{"x": i} for i in range(10)],
            scheduler="processes",
            n_jobs=None,
            arg_type="kwargs",
            progress=True,
            verbose=100,
        )
        assert results == [0, 1, 4, 9, 16, 25, 36, 49, 64, 81]

    def test_tqdm_kwargs(self):
        def fn(x):
            return x**2

        results = dm.parallelized(
            fn,
            [{"x": i} for i in range(10)],
            scheduler="processes",
            n_jobs=None,
            arg_type="kwargs",
            progress=True,
            tqdm_kwargs=dict(desc="My progress bar"),
        )
        assert results == [0, 1, 4, 9, 16, 25, 36, 49, 64, 81]

    def test_with_batch_size(self):
        def _fn(n):
            return n * 3

        def _fn_return_none(n):
            return None

        results = dm.utils.parallelized(
            _fn,
            range(997),
            n_jobs=-1,
            progress=True,
            batch_size=10,
        )
        assert len(results) == 997

        results = dm.utils.parallelized(
            _fn_return_none,
            range(997),
            n_jobs=-1,
            progress=True,
            batch_size=10,
        )
        assert len(results) == 997

    def test_with_total(self):
        def _fn_process_fn(_, row):
            datum = {}
            datum["smiles"] = row["smiles"]
            return pd.Series(datum)

        data = dm.freesolv()
        data = data.iloc[:50]

        # parallel mode

        ## check the `total` arg is ok
        dm.parallelized(
            _fn_process_fn,
            data.iterrows(),
            n_jobs=-1,
            progress=True,
            arg_type="args",
            total=50,
        )

        ## check collision between guessed total and provided one
        dm.parallelized(
            _fn_process_fn,
            list(data.iterrows()),
            n_jobs=-1,
            progress=True,
            arg_type="args",
            total=50,
        )

        # sequential mode

        ## check the `total` arg is ok
        dm.parallelized(
            _fn_process_fn,
            data.iterrows(),
            n_jobs=1,
            progress=True,
            arg_type="args",
            total=50,
        )

        ## check collision between guessed total and provided one
        dm.parallelized(
            _fn_process_fn,
            list(data.iterrows()),
            n_jobs=1,
            progress=True,
            arg_type="args",
            total=50,
        )


def test_parallelized_with_batches():
    data = dm.freesolv()
    data = data.iloc[:10]

    def _fn1(smiles):
        return len(smiles)

    results1 = dm.parallelized(
        _fn1,
        data["smiles"],
        progress=False,
        n_jobs=-1,
    )

    def _fn2(smiles_list):
        return [len(s) for s in smiles_list]

    results2 = dm.parallelized_with_batches(
        _fn2,
        data["smiles"],
        batch_size=2,
        progress=False,
        n_jobs=-1,
    )

    assert results1 == results2



from typing import Union
from typing import List
from typing import Sequence
from typing import Optional
from typing import Tuple

import copy

from loguru import logger

import numpy as np

from rdkit.Chem import AllChem
from rdkit.Chem import rdDistGeom
from rdkit.Chem import rdMolAlign
from rdkit.Chem import rdMolDescriptors
from rdkit.Chem import rdMolTransforms
from rdkit.Chem import rdForceFieldHelpers
from rdkit.ML.Cluster import Butina

from .. import Mol
from .. import mol as dm_mol
from .. import convert
from .. import descriptors


def generate(
    mol: Mol,
    n_confs: Optional[int] = None,
    use_random_coords: bool = True,
    enforce_chirality: bool = True,
    num_threads: int = 1,
    rms_cutoff: Optional[float] = None,
    clear_existing: bool = True,
    align_conformers: bool = True,
    minimize_energy: bool = False,
    sort_by_energy: bool = True,
    method: Optional[str] = None,
    forcefield: str = "UFF",
    ewindow: float = np.inf,
    eratio: float = np.inf,
    energy_iterations: int = 200,
    warning_not_converged: int = 0,
    random_seed: int = 19,
    add_hs: bool = True,
    ignore_failure: bool = False,
    embed_params: Optional[dict] = None,
    verbose: bool = False,
) -> Mol:
    """Compute conformers of a molecule.

    Example:

    ```python
    import datamol as dm
    smiles = "O=C(C)Oc1ccccc1C(=O)O"
    mol = dm.to_mol(smiles)
    mol = dm.conformers.generate(mol)

    # Get all conformers as a list
    conformers = mol.GetConformers()

    # Get the 3D atom positions of the first conformer
    positions = mol.GetConformer(0).GetPositions()

    # If minimization has been enabled (default to True)
    # you can access the computed energy.
    conf = mol.GetConformer(1)
    props = conf.GetPropsAsDict()
    print(props)
    # {'rdkit_UFF_energy': 35.64074017773132,'rdkit_UFF_delta_energy': 0.24682258222552633}
    ```

    Args:
        mol: a molecule
        n_confs: Number of conformers to generate. Depends on the number of rotatable bonds
            by default: 50 for <8, 200 for <12 and 300 for >12.
        use_random_coords: Start the embedding from random coordinates instead of using eigenvalues
            of the distance matrix.
        enforce_chirality: Enforce correct chirilaty if chiral centers are present.
        num_threads: Number of threads to use when embedding multiple conformations.
        rms_cutoff: The minimum RMS value in Angstrom at which two conformers
            are considered redundant and one is deleted. If None, all conformers
            are kept. This step is done after an eventual minimization step.
        clear_existing: Whether to overwrite existing conformers for the molecule.
        align_conformers: Whether to align the conformers.
        minimize_energy: Whether to minimize conformer's energies using MMFF94s.
            Disable to generate conformers much faster.
        sort_by_energy: Sort conformers by energy when minimizing is turned to False.
        method: RDKit method to use for embedding. Choose among
            ["ETDG", "ETKDG", "ETKDGv2", "ETKDGv3"]. If None, "ETKDGv3" is used.
        forcefield: molecular forcefield to use, one of ['UFF','MMFF94s','MMFF94s_noEstat']
        ewindow: maximum energy above minimum energy conformer to output
        eratio: max delta-energy divided by rotatable bonds for conformers
        energy_iterations: Maximum number of iterations during the energy minimization procedure.
            It corresponds to the `maxIters` argument in RDKit.
        warning_not_converged: Wether to log a warning when the number of not converged conformers
            during the minimization is higher than `warning_not_converged`. Only works when `verbose` is set to True.
            Disable with 0. Defaults to 10.
        random_seed: Set to None or -1 to disable.
        add_hs: Whether to add hydrogens to the mol before embedding. If set to True, the hydrogens
            are removed in the returned molecule. Warning: explicit hydrogens won't be conserved. It is strongly
            recommended to let the default value to True. The RDKit documentation says: "To get good 3D conformations,
            it's almost always a good idea to add hydrogens to the molecule first."
        ignore_failure: It set to True, this will avoid raising an error when the embedding fails and return None instead.
        embed_params: Allows the user to specify arbitrary embedding parameters for the conformers. This will override any
            other default settings. See https://www.rdkit.org/docs/source/rdkit.Chem.rdDistGeom.html#rdkit.Chem.rdDistGeom.EmbedParameters
            for more details.
        verbose: Wether to enable logs during the process.

    Returns:
        mol: the molecule with the conformers.
    """

    AVAILABLE_METHODS = ["ETDG", "ETKDG", "ETKDGv2", "ETKDGv3"]

    if method is None:
        method = "ETKDGv3"

    if method not in AVAILABLE_METHODS:
        raise ValueError(f"The method {method} is not supported. Use from {AVAILABLE_METHODS}")

    # Random seed
    if random_seed is None:
        random_seed = -1

    # Clone molecule
    mol = copy.deepcopy(mol)

    # Remove existing conformers
    if clear_existing:
        mol.RemoveAllConformers()

    # Add hydrogens
    if add_hs:
        mol = dm_mol.add_hs(mol)

    rotatable_bonds = descriptors.n_rotatable_bonds(mol)
    if not n_confs:
        # Set the number of conformers depends on
        # the number of rotatable bonds.
        if rotatable_bonds < 8:
            n_confs = 50
        elif rotatable_bonds < 12:
            n_confs = 200
        else:
            n_confs = 300

    # Setup the parameters for the embedding
    params = getattr(rdDistGeom, method)()
    params.randomSeed = random_seed
    params.enforceChirality = enforce_chirality
    params.useRandomCoords = use_random_coords
    params.numThreads = num_threads

    if embed_params is not None:
        for k, v in embed_params.items():
            setattr(params, k, v)

    # Embed conformers
    confs = rdDistGeom.EmbedMultipleConfs(mol, numConfs=n_confs, params=params)

    if len(confs) == 0:
        if ignore_failure:
            if verbose:
                logger.warning(
                    f"Conformers embedding failed for {convert.to_smiles(mol)}. Returning None because ignore_failure is set."
                )
            return None
        raise ValueError(f"Conformers embedding failed for {convert.to_smiles(mol)}")

    energies = None

    # Minimize energy
    if minimize_energy:
        # Minimize conformer's energy using MMFF
        ff = _get_ff(mol, forcefield)
        results = rdForceFieldHelpers.OptimizeMoleculeConfs(
            mol, ff, maxIters=energy_iterations, numThreads=num_threads
        )
        energies = [energy for _, energy in results]

        # Some conformers might not have converged during minimization.
        not_converged = sum([not_converged for not_converged, _ in results if not_converged])
        if warning_not_converged != 0 and not_converged > warning_not_converged and verbose:
            logger.warning(
                f"{not_converged}/{len(results)} conformers have not converged for {convert.to_smiles(mol)}"
            )

    elif sort_by_energy:
        energies = []
        for conf in mol.GetConformers():
            ff = _get_ff(mol, forcefield, conf_id=conf.GetId())
            energies.append(ff.CalcEnergy())
        energies = np.array(energies)

    if energies is not None:
        minE = np.min(energies)
        # Add the energy as a property to each conformers
        [
            (
                conf.SetDoubleProp(f"rdkit_{forcefield}_energy", energy),
                conf.SetDoubleProp(f"rdkit_{forcefield}_delta_energy", energy - minE),
            )
            for energy, conf in zip(energies, mol.GetConformers())
        ]

        # Now we reorder conformers according to their energies,
        # so the lowest energies conformers are first.  eliminate conformers that exceed ewindow, eratio
        mol_clone = copy.deepcopy(mol)
        ordered_conformers = [
            conf
            for E, conf in sorted(zip(energies, mol_clone.GetConformers()), key=lambda x: x[0])
            if E - minE <= ewindow and (E - minE) / rotatable_bonds <= eratio
        ]
        mol.RemoveAllConformers()
        [mol.AddConformer(conf, assignId=True) for conf in ordered_conformers]

    # Align conformers to each others
    if align_conformers:
        rdMolAlign.AlignMolConformers(mol)

    if rms_cutoff is not None:
        mol = cluster(
            mol,
            rms_cutoff=rms_cutoff,
            already_aligned=align_conformers,
            centroids=True,
        )  # type: ignore

    if add_hs:
        mol = dm_mol.remove_hs(mol)

    return mol


def _get_ff(mol: Mol, forcefield: str, conf_id: int = -1):
    """Gets molecular forcefield for input mol according to name
    Args:
        mol: input molecule
        forcefield: forcefield name. One of "UFF", "MMFF94s", "MMFF94s_noEstat"]
        conf_id: conformer id. -1 is used by default
    """
    assert forcefield in [
        "UFF",
        "MMFF94s",
        "MMFF94s_noEstat",
    ], f"Forcefield {forcefield} is not supported"
    if forcefield == "UFF":
        return rdForceFieldHelpers.UFFGetMoleculeForceField(mol, confId=conf_id)

    mp = rdForceFieldHelpers.MMFFGetMoleculeProperties(mol, "MMFF94s")
    if forcefield == "MMFF94s_noEstat":
        mp.SetMMFFEleTerm(False)
    return rdForceFieldHelpers.MMFFGetMoleculeForceField(mol, mp, confId=conf_id)


def cluster(
    mol: Mol,
    rms_cutoff: float = 1,
    already_aligned: bool = False,
    centroids: bool = True,
):
    """Cluster the conformers of a molecule according to an RMS threshold in Angstrom.

    Args:
        mol: a molecule
        rms_cutoff: The RMS cutoff in Angstrom.
        already_aligned: Whether or not the conformers are aligned. If False,
            they will be aligmned furing the RMS computation.
        centroids: If True, return one molecule with centroid conformers
            only. If False return a list of molecules per cluster with all
            the conformers of the cluster. Defaults to True.
    """

    # Clone molecule
    mol = copy.deepcopy(mol)

    # Compute RMS
    dmat = AllChem.GetConformerRMSMatrix(mol, prealigned=already_aligned)

    # Cluster
    conf_clusters = Butina.ClusterData(
        dmat,
        nPts=mol.GetNumConformers(),
        distThresh=rms_cutoff,
        isDistData=True,
        reordering=False,
    )

    return return_centroids(mol, conf_clusters, centroids=centroids)


def rmsd(mol: Mol) -> np.ndarray:
    """Compute the RMSD between all the conformers of a molecule.

    Args:
        mol: a molecule
    """

    if mol.GetNumConformers() <= 1:
        raise ValueError(
            "The molecule has 0 or 1 conformer. You can generate conformers with `dm.conformers.generate(mol)`."
        )

    n_confs = mol.GetNumConformers()
    rmsds = []
    for i in range(n_confs):
        for j in range(n_confs):
            rmsd = rdMolAlign.AlignMol(prbMol=mol, refMol=mol, prbCid=i, refCid=j)
            rmsds.append(rmsd)
    return np.array(rmsds).reshape(n_confs, n_confs)


def return_centroids(
    mol: Mol,
    conf_clusters: Sequence[Sequence[int]],
    centroids: bool = True,
) -> Union[List[Mol], Mol]:
    """Given a list of cluster indices, return one single molecule
    with only the centroid of the clusters of a list of molecules per cluster.

    Args:
        mol: a molecule.
        conf_clusters: list of cluster indices.
        centroids: If True, return one molecule with centroid conformers
            only. If False return a list of molecules per cluster with all
            the conformers of the cluster.
    """

    if centroids:
        # Collect centroid of each cluster (first element of the list)
        centroid_list = [indices[0] for indices in conf_clusters]

        # Keep only centroid conformers
        mol_clone = copy.deepcopy(mol)
        confs = [mol_clone.GetConformers()[i] for i in centroid_list]
        mol.RemoveAllConformers()
        [mol.AddConformer(conf, assignId=True) for conf in confs]
        return mol

    else:
        # Create a new molecule for each cluster and add conformers to it.
        mols = []
        for cluster in conf_clusters:
            m = copy.deepcopy(mol)
            m.RemoveAllConformers()
            [m.AddConformer(mol.GetConformer(c), assignId=True) for c in cluster]
            mols.append(m)
        return mols


def translate(mol: Mol, new_centroid: Union[np.ndarray, List[int]], conf_id: int = -1):
    """Move a given conformer of a molecule to a new position. The transformation is performed
    in place.

    Args:
        mol: the molecule.
        new_centroid: the new position to move to of shape [x, y, z]
        conf_id: id of the conformer.
    """

    # Get conformer
    conf = mol.GetConformer(conf_id)

    # Compute the vector for translation
    mol_center = rdMolTransforms.ComputeCentroid(conf)
    mol_center = np.array([mol_center.x, mol_center.y, mol_center.z])

    # Make the transformation matrix
    T = np.eye(4)
    T[:3, 3] = new_centroid - mol_center

    # Transform
    rdMolTransforms.TransformConformer(conf, T)


def align_conformers(
    mols: List[Mol],
    ref_id: int = 0,
    copy: bool = True,
    conformer_id: int = -1,
    backend: str = "crippenO3A",
) -> Tuple[list, list]:
    """Align a list of molecules to a reference molecule.

    Note that using the `O3A` backend, hydrogens will be added at the beginning of the procedure
    and removed at the end of the procedure.

    Args:
        mols: List of molecules to align. All the molecules must have a conformer.
        ref_id: Index of the reference molecule. By default, the first molecule in the list
            will be used as reference.
        copy: Whether to copy the molecules before performing the alignement.
        conformer_id: Conformer id to use.
        backend: Backend to use to compute the alignment from `crippenO3A`, `O3A`.

    Returns:
        mols: The aligned molecules.
        scores: The score of the alignement.
    """

    allowed_backends = ["crippenO3A", "O3A"]
    if backend not in allowed_backends:
        raise ValueError(
            f"The backend '{backend}' is not supported. Choose from: {allowed_backends}"
        )

    # Check all input molecules has a conformer
    if not all([mol.GetNumConformers() >= 1 for mol in mols]):
        raise ValueError("One or more input molecules is missing a conformer.")

    # Make a copy of the molecules since they are going to be modified
    if copy:
        mols = [dm_mol.copy_mol(mol) for mol in mols]

    # Split ref and probe mols
    mol_ref = mols[ref_id]
    mol_probes = mols

    if backend == "crippenO3A":
        # Compute Crippen contributions for every atoms and molecules
        crippen_contribs = [rdMolDescriptors._CalcCrippenContribs(mol) for mol in mol_probes]
        crippen_contrib_ref = crippen_contribs[ref_id]
        crippen_contrib_probes = crippen_contribs

        # Loop and align
        # NOTE(hadim): we could eventually parallelize this if that's needed.

        scores = []
        for i, mol in enumerate(mol_probes):
            crippenO3A = rdMolAlign.GetCrippenO3A(
                prbMol=mol,
                refMol=mol_ref,
                prbCrippenContribs=crippen_contrib_probes[i],
                refCrippenContribs=crippen_contrib_ref,
                prbCid=conformer_id,
                refCid=conformer_id,
                maxIters=50,
            )
            crippenO3A.Align()

            scores.append(crippenO3A.Score())

    elif backend == "O3A":
        # Add hydrogens first
        mol_probes = [dm_mol.add_hs(mol, add_coords=True) for mol in mol_probes]
        mol_ref = dm_mol.add_hs(mol_ref, add_coords=True)

        # Compute MMFF params for every molecules
        mmff_params = [rdForceFieldHelpers.MMFFGetMoleculeProperties(mol) for mol in mol_probes]

        # Split reference and probe molecules
        mmff_params_ref = mmff_params[ref_id]
        mmff_params_probes = mmff_params

        # Loop and align
        # NOTE(hadim): we could eventually parallelize this if that's needed.

        scores = []
        for i, mol in enumerate(mol_probes):
            pyO3A = rdMolAlign.GetO3A(
                prbMol=mol,
                refMol=mol_ref,
                prbPyMMFFMolProperties=mmff_params_probes[i],
                refPyMMFFMolProperties=mmff_params_ref,
                prbCid=conformer_id,
                refCid=conformer_id,
                maxIters=50,
            )
            pyO3A.Align()

            scores.append(pyO3A.Score())

        # Remove the hydrogens
        mol_probes = [dm_mol.remove_hs(mol) for mol in mol_probes]

    else:
        raise ValueError(f"Backend {backend} not supported.")

    scores = np.array(scores)

    return mol_probes, scores
